from importlib.metadata import version, PackageNotFoundError
from maiconverter.cli import main

try:
    __version__ = version("maiconverter")
except PackageNotFoundError:
    __version__ = "not installed"



from __future__ import annotations

import math
from typing import Optional, Tuple, List, Union
from lark import Lark

from .tools import (
    get_measure_divisor,
    convert_to_fragment,
    get_rest,
    parallel_parse_fragments,
)
from ..event import NoteType
from .simainote import TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote, BPM
from .simai_parser import SimaiTransformer

# I hate the simai format can we use bmson or stepmania chart format for
# community-made charts instead
from ..tool import measure_to_second, second_to_measure, offset_arg_to_measure


class SimaiChart:
    """A class that represents a simai chart. Contains notes and bpm
    information. Does not include information such as
    song name, chart difficulty, composer, chart maker, etc.
    It only contains enough information to build a working simai
    chart.

    Attributes:
        bpms: Contains bpm events of the chart.
        notes: Contains notes of the chart.
    """

    def __init__(self):
        self.notes: List[
            Union[TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote]
        ] = []
        self.bpms: List[BPM] = []
        self._divisor: Optional[float] = None
        self._measure = 1.0

    @classmethod
    def from_str(cls, chart_text: str, message: Optional[str] = None) -> SimaiChart:
        # TODO: Rewrite this
        if message is None:
            print("Parsing simai chart...", end="", flush=True)
        else:
            print(message, end="", flush=True)

        simai_chart = cls()
        chart_text = "".join(chart_text.split())
        try:
            events_list = parallel_parse_fragments(chart_text.split(","))
        except:
            print("ERROR")
            raise
        else:
            print("Done")

        for events in events_list:
            star_positions = []
            offset = 0
            for event in events:
                event_type = event["type"]
                if event_type == "bpm":
                    simai_chart.set_bpm(simai_chart._measure, event["value"])
                elif event_type == "divisor":
                    simai_chart._divisor = event["value"]
                elif event_type == "tap":
                    is_break, is_ex, is_star = False, False, False
                    modifier = event["modifier"]
                    if "b" in modifier:
                        is_break = True
                    if "x" in modifier:
                        is_ex = True
                    if "$" in modifier:
                        is_star = True

                    if "`" in modifier:
                        # Equivalent to one tick in ma2 with resolution of 384
                        offset += 0.0027
                    else:
                        offset = 0

                    simai_chart.add_tap(
                        measure=simai_chart._measure + offset,
                        position=event["button"],
                        is_break=is_break,
                        is_star=is_star,
                        is_ex=is_ex,
                    )
                elif event_type == "hold":
                    is_ex = False
                    modifier = event["modifier"]
                    if "x" in modifier:
                        is_ex = True

                    if "`" in modifier:
                        # Equivalent to one tick in ma2 with resolution of 384
                        offset += 0.0027
                    else:
                        offset = 0

                    simai_chart.add_hold(
                        measure=simai_chart._measure + offset,
                        position=event["button"],
                        duration=event["duration"],
                        is_ex=is_ex,
                    )
                elif event_type == "slide":
                    is_break, is_ex, is_tapless = False, False, False
                    modifier = event["modifier"]
                    if "b" in modifier:
                        is_break = True
                    if "x" in modifier:
                        is_ex = True
                    if any([a in modifier for a in "?!$"]):
                        # Tapless slides
                        # ? means the slide has no tap
                        # ! produces a tapless slide with no path, just a moving star
                        # $ is a remnant of 2simai, it is equivalent to ?
                        is_tapless = True

                    if "*" in modifier:
                        # Chained slides should have the same offset
                        pass
                    elif "`" in modifier:
                        # Equivalent to one tick in ma2 with resolution of 384
                        offset += 0.0027
                    else:
                        offset = 0

                    if not (is_tapless or event["start_button"] in star_positions):
                        simai_chart.add_tap(
                            measure=simai_chart._measure + offset,
                            position=event["start_button"],
                            is_break=is_break,
                            is_star=True,
                            is_ex=is_ex,
                        )
                        star_positions.append(event["start_button"])

                    equivalent_bpm = event["equivalent_bpm"]
                    duration = event["duration"]
                    delay = 0.25
                    if equivalent_bpm is not None:
                        multiplier = (
                            simai_chart.get_bpm(simai_chart._measure) / equivalent_bpm
                        )
                        duration = multiplier * duration
                        delay = multiplier * delay

                    simai_chart.add_slide(
                        measure=simai_chart._measure + offset,
                        start_position=event["start_button"],
                        end_position=event["end_button"],
                        duration=duration,
                        pattern=event["pattern"],
                        delay=delay,
                        reflect_position=event["reflect_position"],
                    )
                elif event_type == "touch_tap":
                    is_firework = False
                    modifier = event["modifier"]
                    if "f" in modifier:
                        is_firework = True

                    if "`" in modifier:
                        # Equivalent to one tick in ma2 with resolution of 384
                        offset += 0.0027
                    else:
                        offset = 0

                    simai_chart.add_touch_tap(
                        measure=simai_chart._measure + offset,
                        position=event["location"],
                        region=event["region"],
                        is_firework=is_firework,
                    )

                elif event_type == "touch_hold":
                    is_firework = False
                    modifier = event["modifier"]
                    if "f" in modifier:
                        is_firework = True

                    if "`" in modifier:
                        # Equivalent to one tick in ma2 with resolution of 384
                        offset += 0.0027
                    else:
                        offset = 0

                    simai_chart.add_touch_hold(
                        measure=simai_chart._measure + offset,
                        position=event["location"],
                        region=event["region"],
                        duration=event["duration"],
                        is_firework=is_firework,
                    )
                else:
                    raise Exception(f"Unknown event type: {event_type}")

            simai_chart._measure += 1 / simai_chart._divisor

        return simai_chart

    @classmethod
    def open(cls, file: str) -> SimaiChart:
        """Opens a text file containing only a Simai chart. Does NOT accept a regular Simai file which contains
        metadata and multiple charts. Use `parse_file` to parse a normal Simai file.

        Args:
              file: The path of the Simai chart file.

        Examples:
            Open a Simai chart file named "example.txt" at current directory.

            >>> simai = SimaiChart.open("./example.txt")
        """
        with open(file, "r") as f:
            chart = f.read()

        return cls.from_str(chart)

    def add_tap(
        self,
        measure: float,
        position: int,
        is_break: bool = False,
        is_star: bool = False,
        is_ex: bool = False,
    ) -> SimaiChart:
        """Adds a tap note to the list of notes.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the tap note happens.
            is_break: Whether a tap note is a break note.
            is_star: Whether a tap note is a star note.
            is_ex: Whether a tap note is an ex note.

        Examples:
            Add a regular tap note at measure 1, break tap note at
            measure 2, ex tap note at measure 2.5, star note at
            measure 3, and a break star note at measure 5. All at
            position 7.

            >>> simai = SimaiChart()
            >>> simai.add_tap(1, 7)
            >>> simai.add_tap(2, 7, is_break=True)
            >>> simai.add_tap(2.5, 7, is_ex=True)
            >>> simai.add_tap(3, 7, is_star=True)
            >>> simai.add_tap(5, 7, is_break=True, is_star=True)
        """
        tap_note = TapNote(
            measure=measure,
            position=position,
            is_break=is_break,
            is_star=is_star,
            is_ex=is_ex,
        )
        self.notes.append(tap_note)

        return self

    def del_tap(self, measure: float, position: int) -> SimaiChart:
        """Deletes a tap note from the list of notes.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the tap note happens.

        Examples:
            Remove tap note at measure 26.75 at button 4.
            >>> simai = SimaiChart()
            >>> simai.add_tap(26.5, 4)
            >>> simai.del_tap(26.75, 4)
        """
        tap_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == position
        ]
        for note in tap_notes:
            self.notes.remove(note)

        return self

    def add_hold(
        self,
        measure: float,
        position: int,
        duration: float,
        is_ex: bool = False,
    ) -> SimaiChart:
        """Adds a hold note to the list of notes.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the hold note happens.
            duration: Total time duration of the hold note.
            is_ex: Whether a hold note is an ex note.

        Examples:
            Add a regular hold note at button 2 at measure 1, with
            duration of 5 measures. And an ex hold note at button
            6 at measure 3, with duration of 0.5 measures.

            >>> simai = SimaiChart()
            >>> simai.add_hold(1, 2, 5)
            >>> simai.add_hold(3, 6, 0.5, is_ex=True)
        """
        hold_note = HoldNote(measure, position, duration, is_ex)
        self.notes.append(hold_note)

        return self

    def del_hold(self, measure: float, position: int) -> SimaiChart:
        """Deletes the matching hold note in the list of notes. If there are multiple
        matches, all matching notes are deleted. If there are no match, nothing happens.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the hold note happens.

        Examples:
            Add a regular hold note at button 0 at measure 3.25 with duration of 2 measures
            and delete it.

            >>> simai = SimaiChart()
            >>> simai.add_hold(3.25, 0, 2)
            >>> simai.del_hold(3.25, 0)
        """
        hold_notes = [
            x
            for x in self.notes
            if isinstance(x, HoldNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == position
        ]
        for note in hold_notes:
            self.notes.remove(note)

        return self

    def add_slide(
        self,
        measure: float,
        start_position: int,
        end_position: int,
        duration: float,
        pattern: str,
        delay: float = 0.25,
        reflect_position: Optional[int] = None,
    ) -> SimaiChart:
        """Adds both a slide note to the list of notes.

        Args:
            measure: Time when the slide starts, in
                terms of measures.
            start_position: Button where the slide starts.
            end_position: Button where the slide ends.
            duration: Total duration of the slide, in terms of
                measures. Includes slide delay.
            pattern: The one or two character slide pattern used.
            delay: Duration from when the slide appears and when it
                starts to move, in terms of measures. Defaults to 0.25.
            reflect_position: The button where the 'V' slide will first go to.
                Optional, defaults to None.

        Examples:
            Add a '-' slide at measure 2.25 from button 1 to button 5 with
            duration of 1.5 measures

            >>> simai = SimaiChart()
            >>> simai.add_slide(2.25, 1, 5, 1.5, "-")
            >>> simai.add_slide(3, 2, 7, 0.5, "V", reflect_position=4)
        """
        slide_note = SlideNote(
            measure,
            start_position,
            end_position,
            duration,
            pattern,
            delay,
            reflect_position,
        )
        self.notes.append(slide_note)

        return self

    def del_slide(
        self,
        measure: float,
        start_position: int,
        end_position: int,
    ) -> SimaiChart:
        slide_notes = [
            x
            for x in self.notes
            if isinstance(x, SlideNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == start_position
            and x.end_position == end_position
        ]
        for note in slide_notes:
            self.notes.remove(note)

        return self

    def add_touch_tap(
        self,
        measure: float,
        position: int,
        region: str,
        is_firework: bool = False,
    ) -> SimaiChart:
        touch_tap_note = TouchTapNote(measure, position, region, is_firework)
        self.notes.append(touch_tap_note)

        return self

    def del_touch_tap(
        self,
        measure: float,
        position: int,
        region: str,
    ) -> SimaiChart:
        touch_taps = [
            x
            for x in self.notes
            if isinstance(x, TouchTapNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == position
            and x.region == region
        ]
        for note in touch_taps:
            self.notes.remove(note)

        return self

    def add_touch_hold(
        self,
        measure: float,
        position: int,
        region: str,
        duration: float,
        is_firework: bool = False,
    ) -> SimaiChart:
        touch_hold_note = TouchHoldNote(
            measure, position, region, duration, is_firework
        )
        self.notes.append(touch_hold_note)

        return self

    def del_touch_hold(
        self,
        measure: float,
        position: int,
        region: str,
    ) -> SimaiChart:
        touch_holds = [
            x
            for x in self.notes
            if isinstance(x, TouchHoldNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == position
            and x.region == region
        ]
        for note in touch_holds:
            self.notes.remove(note)

        return self

    def set_bpm(self, measure: float, bpm: float) -> SimaiChart:
        """Sets the bpm at given measure.

        Note:
            If BPM event is already defined at given measure,
            the method will overwrite it.

        Args:
            measure: Time, in measures, where the bpm is defined.
            bpm: The tempo in beat per minutes.

        Examples:
            In a chart, the initial bpm is 180 then changes
            to 250 in measure 12.

            >>> simai = SimaiChart()
            >>> simai.set_bpm(0, 180)
            >>> simai.set_bpm(12, 250)
        """
        self.del_bpm(measure)

        bpm_event = BPM(measure, bpm)
        self.bpms.append(bpm_event)

        return self

    def get_bpm(self, measure: float) -> float:
        """Gets the bpm at given measure.

        Args:
            measure: Time, in measures.

        Returns:
            Returns the bpm defined at given measure or None.

        Raises:
            ValueError: When measure is negative, there are no BPMs
                defined, or there are no starting BPM defined.

        Examples:
            In a chart, the initial bpm is 180 then changes
            to 250 in measure 12.

            >>> simai = SimaiChart()
            >>> simai.get_bpm(0)
            180.0
            >>> simai.get_bpm(11.99)
            180.0
            >>> simai.get_bpm(12)
            250.0
        """
        if len(self.bpms) == 0:
            raise ValueError("No BPMs defined")
        if not any([0.0 <= x.measure <= 1.0 for x in self.bpms]):
            raise ValueError("No starting BPM defined")

        self.bpms.sort(key=lambda x: x.measure)
        previous_bpm = self.bpms[0].bpm
        for bpm in self.bpms:
            if math.isclose(measure, bpm.measure, abs_tol=0.0001):
                return bpm.bpm
            if bpm.measure > measure:
                break

            previous_bpm = bpm.bpm

        return previous_bpm

    def del_bpm(self, measure: float) -> SimaiChart:
        """Deletes the bpm at given measure.

        Note:
            If there are no BPM defined for that measure, nothing happens.

        Args:
            measure: Time, in measures, where the bpm is defined.

        Examples:
            Delete the BPM change defined at measure 24.

            >>> simai = SimaiChart()
            >>> simai.del_bpm(24)
        """
        bpms = [
            x for x in self.bpms if math.isclose(x.measure, measure, abs_tol=0.0001)
        ]
        for x in bpms:
            self.bpms.remove(x)

        return self

    def offset(self, offset: Union[float, str]) -> SimaiChart:
        offset = offset_arg_to_measure(offset, self.second_to_measure)

        for note in self.notes:
            note.measure = round(note.measure + offset, 4)

        for bpm in self.bpms:
            if 0 <= bpm.measure <= 1:
                continue

            bpm.measure = round(bpm.measure + offset, 4)

        return self

    def measure_to_second(self, measure: float) -> float:
        bpms = [(bpm.measure, bpm.bpm) for bpm in self.bpms]

        return measure_to_second(measure, bpms)

    def second_to_measure(self, seconds: float) -> float:
        bpms = [(bpm.measure, bpm.bpm) for bpm in self.bpms]
        measure = second_to_measure(seconds, bpms)

        return measure

    def export(self, max_den: int = 1000) -> str:
        # TODO: Rewrite this
        measures = [event.measure for event in self.notes + self.bpms]

        measures += [int(i) for i in measures]
        measures.append(0.0)

        measures = list(set(measures))
        last_whole_measure = max([int(measure) for measure in measures])
        measures.sort()

        # whole_divisors contains divisors that fit perfectly all notes in one measure.
        # It either contains an integer or None.
        whole_divisors: List[Union[int, None]] = []
        for whole_measure in range(last_whole_measure + 1):
            note_measures = [
                note_measure
                for note_measure in measures
                if int(note_measure) == whole_measure
            ]
            whole_divisors.append(get_measure_divisor(note_measures))

        # last_measure takes into account slide and hold notes' end measure
        last_measure = 0.0
        # measure_tick is our time-tracking variable. Used to know what measure
        # are we in-between rests ","
        measure_tick = 0.0
        # previous_divisor is used for comparing to current_divisor
        # to know if we should add a "{}" indicator
        previous_divisor: Optional[int] = None
        # previous_measure_int is used for comparing to current measure.
        # If we are in a new whole measure, add a new line and add the divisor.
        previous_measure_int = 0
        # Our resulting chart in text form. Assuming that string fits in memory
        result = ""
        for (i, current_measure) in enumerate(measures):
            bpm = [bpm for bpm in self.bpms if bpm.measure == current_measure]
            notes = [note for note in self.notes if note.measure == current_measure]

            hold_slides = [
                note
                for note in notes
                if note.note_type
                in [
                    NoteType.hold,
                    NoteType.ex_hold,
                    NoteType.touch_hold,
                    NoteType.complete_slide,
                ]
            ]
            for hold_slide in hold_slides:
                # Get hold and slide end measure and compare with last_measure
                if hold_slide.note_type == NoteType.complete_slide:
                    last_measure = max(
                        current_measure + hold_slide.delay + hold_slide.duration,
                        last_measure,
                    )
                else:
                    last_measure = max(
                        current_measure + hold_slide.duration, last_measure
                    )

            whole_divisor = whole_divisors[int(current_measure)]

            if i == len(measures) - 1:
                # We are at the end so let's check if there are any
                # active holds or slides
                if last_measure > current_measure:
                    (whole, current_divisor, rest_amount) = get_rest(
                        current_measure,
                        last_measure,
                        current_divisor=(
                            previous_divisor if whole_divisor is None else whole_divisor
                        ),
                        max_den=max_den,
                    )

                else:
                    # Nothing to do
                    current_divisor = (
                        previous_divisor if whole_divisor is None else whole_divisor
                    )
                    whole, rest_amount = 0, 0
            else:
                # Why doesn't Python have a safe list 'get' method
                next_measure: Optional[float] = (
                    measures[i + 1] if i + 1 < len(measures) else None
                )
                after_next_measure: Optional[float] = (
                    measures[i + 2] if i + 2 < len(measures) else None
                )
                (whole, current_divisor, rest_amount) = get_rest(
                    current_measure,
                    next_measure,
                    after_next_measure=after_next_measure,
                    current_divisor=(
                        previous_divisor if whole_divisor is None else whole_divisor
                    ),
                    max_den=max_den,
                )

            if (
                previous_divisor != current_divisor
                or int(measure_tick) > previous_measure_int
            ):
                result += "\n"
                result += convert_to_fragment(
                    notes + bpm,
                    self.get_bpm(current_measure + 1),
                    current_divisor,
                    max_den=max_den,
                )
                previous_divisor = current_divisor
                previous_measure_int = int(measure_tick)
            else:
                result += convert_to_fragment(
                    notes + bpm, self.get_bpm(current_measure + 1), max_den=max_den
                )

            measure_tick = current_measure

            for _ in range(rest_amount):
                result += ","
                measure_tick += 1 / current_divisor

            if whole > 0:
                if current_divisor != 1:
                    result += "{1}"
                    previous_divisor = 1

                for _ in range(whole):
                    result += ","
                    measure_tick += 1

            measure_tick = round(measure_tick * 10000) / 10000

        result += ",\nE\n"
        return result


def parse_file_str(
    file: str, lark_file: str = "simai.lark"
) -> Tuple[str, List[Tuple[int, SimaiChart]]]:
    parser = Lark.open(lark_file, rel_to=__file__, parser="lalr")

    dicts: List[dict] = SimaiTransformer().transform(parser.parse(file))

    title = ""
    charts: List[Tuple[int, SimaiChart]] = []
    for element in dicts:
        if element["type"] == "title":
            title: str = element["value"]
        elif element["type"] == "chart":
            num, chart = element["value"]
            simai_chart = SimaiChart.from_str(chart, message=f"Parsing chart #{num}...")
            charts.append((num, simai_chart))

    return title, charts


def parse_file(
    path: str,
    encoding: str = "UTF-8",
    lark_file: str = "simai.lark",
) -> Tuple[str, List[Tuple[int, SimaiChart]]]:
    with open(path, encoding=encoding) as f:
        simai = f.read()

    print(f"Parsing Simai file at {path}")
    try:
        result = parse_file_str(simai, lark_file=lark_file)
    except:
        print(f"Error parsing Simai file at {path}")
        raise
    else:
        print(f"Done parsing Simai file at {path}")
        return result



from .simai_parser import *
from .simainote import *
from .tools import (
    get_rest,
    convert_to_fragment,
    get_measure_divisor,
    handle_tap,
    handle_hold,
    handle_slide,
    handle_touch_tap,
    handle_touch_hold,
)
from .simai import SimaiChart, parse_file, parse_file_str



import enum
import math
from functools import total_ordering
from typing import Dict

from .event import Event, EventType

# Note types
# 0: Start slide (For SDT)
# 1: Regular Tap
# 2: Hold
# 3: Break tap
# 4: Star
# 5: Break star
# 6: Ex tap
# 7: Ex star
# 8: Ex hold
# 9: Touch tap
# 10: Touch hold
# 11: Complete slide (Both start and end slide. For ma2 and simai)
# 128: End slide (For SDT)


class NoteType(enum.Enum):
    start_slide = 0
    tap = 1
    hold = 2
    break_tap = 3
    star = 4
    break_star = 5
    ex_tap = 6
    ex_star = 7
    ex_hold = 8
    touch_tap = 9
    touch_hold = 10
    complete_slide = 11
    end_slide = 128


NOTE_ORDERING: Dict[int, int] = {
    5: 0,
    4: 0,
}


class Note(Event):
    def __init__(self, measure: float, position: int, note_type: NoteType) -> None:
        if position < 0:
            raise ValueError("Note position is negative " + str(position))

        super().__init__(measure, EventType.note)
        self.position = position
        self.note_type = note_type

    def __key(self):
        return self.measure, self.event_type, self.position, self.note_type


@total_ordering
class MaiNote(Note):
    def __hash__(self):
        return hash(self.__key())

    def __lt__(self, other):
        if math.isclose(self.measure, other.measure, abs_tol=0.0001):
            if self.position == other.position:
                return self.note_type.value < other.note_type.value
            return self.position < other.position
        return self.measure < other.measure

    def __eq__(self, other):
        return (
            math.isclose(self.measure, other.measure, abs_tol=0.0001)
            and self.position == other.position
            and self.note_type.value == other.note_type.value
        )


@total_ordering
class SimaiNote(Note):
    def __hash__(self):
        return hash(self.__key())

    def __lt__(self, other):
        if math.isclose(self.measure, other.measure, abs_tol=0.0001):
            if self.note_type == other.note_type:
                return self.position < other.position
            else:
                return self.note_type.value < other.note_type.value
        else:
            return self.measure < other.measure

    def __eq__(self, other):
        return (
            math.isclose(self.measure, other.measure, abs_tol=0.0001)
            and self.note_type.value == other.note_type.value
            and self.position == other.position
        )



import math
from typing import Tuple

from maiconverter.event import MaiNote, NoteType, Event, EventType
from maiconverter.tool import slide_distance

# Dictionary for a note type's representation in ma2
# Does not cover slide notes, BPM, and meter events.
# Use slide_dict for slides instead.
note_dict = {
    "TAP": 1,
    "HLD": 2,
    "BRK": 3,
    "STR": 4,
    "BST": 5,
    "XTP": 6,
    "XST": 7,
    "XHO": 8,
    "TTP": 9,
    "THO": 10,
    # "SLD": 11
}

# Dictionary for a slide note's representation in ma2 by pattern.
slide_dict = {
    "SI_": 1,
    "SCL": 2,
    "SCR": 3,
    "SUL": 4,
    "SUR": 5,
    "SSL": 6,
    "SSR": 7,
    "SV_": 8,
    "SXL": 9,
    "SXR": 10,
    "SLL": 11,
    "SLR": 12,
    "SF_": 13,
}


class SlideNote(MaiNote):
    def __init__(
        self,
        measure: float,
        start_position: int,
        end_position: int,
        pattern: int,
        duration: float,
        delay: float = 0.25,
    ) -> None:
        """Produces a ma2 slide note.

        Note:
            Please use MaiMa2 class' add_slide method for adding slides.

        Args:
            measure: Time when the note starts, in terms of measures.
            start_position: Starting button.
            end_position: Ending button.
            pattern: Numerical representation of the slide pattern.
            duration: Time duration of when the slide starts moving and
                      when it ends (delay is not included.) in terms of
                      measures.
            delay: Time duration of when the slide appears and when it
                   starts to move, in terms of measures.
        Raises:
            ValueError: When pattern or duration is not a positive integer,
                        and when delay, or end_position are negative.
        """
        if pattern <= 0:
            raise ValueError(f"Slide pattern is not positive {pattern}")
        if duration <= 0:
            raise ValueError(f"Slide duration is not positive {duration}")
        if delay < 0:
            raise ValueError(f"Slide delay is negative {delay}")
        if end_position < 0:
            raise ValueError(f"Slide end position is negative {end_position}")

        super().__init__(measure, start_position, NoteType.complete_slide)
        self.end_position = end_position
        self.pattern = pattern
        self.delay = delay
        self.duration = duration

    def to_str(self, resolution: int = 384) -> str:
        measure = measure_to_ma2_time(self.measure, resolution)
        template = "{}\t{}\t{}\t{}\t{}\t{}\t{}"
        inv_slide_dict = {v: k for k, v in slide_dict.items()}
        if self.pattern not in inv_slide_dict:
            raise ValueError(f"Unknown slide pattern {self.pattern}")

        pattern = inv_slide_dict[self.pattern]
        delay = round(self.delay * resolution)
        duration = round(self.duration * resolution)
        return template.format(
            pattern,
            measure[0],
            measure[1],
            self.position,
            delay,
            duration,
            self.end_position,
        )


class HoldNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        duration: float,
        is_ex: bool = False,
    ) -> None:
        """Produces a ma2 hold note.

        Note:
            Please use MaiMa2 class' add_hold method for adding holds.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the hold note happens.
            duration: Total time duration of the hold note.
            is_ex: Whether a hold note is an ex note.

        Raises:
            ValueError: When duration is not positive.
        """
        if duration < 0:
            raise ValueError(f"Hold duration is negative: {duration}")

        if is_ex:
            super().__init__(measure, position, NoteType.ex_hold)
        else:
            super().__init__(measure, position, NoteType.hold)

        self.duration = duration

    def to_str(self, resolution: int) -> str:
        measure = measure_to_ma2_time(self.measure, resolution)
        template = "HLD\t{}\t{}\t{}\t{}"
        duration = round(self.duration * resolution)
        return template.format(measure[0], measure[1], self.position, duration)


class TapNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        is_star: bool = False,
        is_break: bool = False,
        is_ex: bool = False,
    ) -> None:
        """Produces a ma2 tap note.

        Note:
            Please use MaiMa2 class' add_tap method for adding taps.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the tap note happens.
            is_star: Whether a tap note is a star note.
            is_break: Whether a tap note is a break note.
            is_ex: Whether a tap note is an ex note.
        """

        if is_ex and is_star:
            super().__init__(measure, position, NoteType.ex_star)
        elif is_ex and not is_star:
            super().__init__(measure, position, NoteType.ex_tap)
        elif is_star and is_break:
            super().__init__(measure, position, NoteType.break_star)
        elif is_star and not is_break:
            super().__init__(measure, position, NoteType.star)
        elif not is_star and is_break:
            super().__init__(measure, position, NoteType.break_tap)
        elif not is_star and not is_break:
            super().__init__(measure, position, NoteType.tap)

    def to_str(self, resolution: int) -> str:
        measure = measure_to_ma2_time(self.measure, resolution)
        template = "{}\t{}\t{}\t{}"
        inv_note_dict = {v: k for k, v in note_dict.items()}
        if self.note_type.value not in inv_note_dict:
            raise ValueError(f"Unknown tap note {self.note_type.value}")

        name = inv_note_dict[self.note_type.value]
        return template.format(name, measure[0], measure[1], self.position)


class TouchTapNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        region: str,
        is_firework: bool = False,
        size: str = "M1",
    ) -> None:
        """Produces a ma2 touch tap note.

        Note:
            Please use MaiMa2 class' add_touch_tap method
            for adding touch taps.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Position in the region where the note happens.
            region: Touch region where the note happens.
            is_firework: Whether a touch tap note will produce
                         fireworks. Optional bool, defaults to False.
            size: Optional str. Specifies the size of touch note defaults to "M1"

        Examples:
            A touch tap note at E0 at measure 2.25, produces no
            fireworks.

            >>> touch_tap = TouchTapNote(2.25, 0, "E")

            A touch tap note at B5 at measure 5.00, produces fireworks.

            >>> touch_tap = TouchTapNote(5.00, 5, "B", True)
        """
        if size not in ["M1", "L1"]:
            raise ValueError(f"Invalid size given: {size}")

        super().__init__(measure, position, NoteType.touch_tap)
        self.is_firework = is_firework
        self.region = region
        self.size = size

    def to_str(self, resolution: int) -> str:
        measure = measure_to_ma2_time(self.measure, resolution)
        template = "TTP\t{}\t{}\t{}\t{}\t{}\t{}"
        fireworks = 1 if self.is_firework else 0
        return template.format(
            measure[0],
            measure[1],
            self.position,
            self.region,
            fireworks,
            self.size,
        )


class TouchHoldNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        region: str,
        duration: float,
        is_firework: bool = False,
        size: str = "M1",
    ) -> None:
        """Produces a ma2 touch hold note.

        Note:
            Please use MaiMa2 class' add_touch_hold method
            for adding touch holds.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Position in the region where the note happens.
            region: Touch region where the note happens.
            duration: Total time duration of the touch hold note.
            is_firework: Whether a touch hold note will produce
                         fireworks. Optional bool, defaults to False.
            size: Optional str. Specifies the size of touch note defaults to "M1"

        Examples:
            A touch hold note at C0 at measure 1 with duration of
            1.5 measures, produces no fireworks.

            >>> touch_tap = TouchHoldNote(1, 0, "C", 1.5)
        """
        if duration < 0:
            raise ValueError(f"Hold duration is negative: {duration} ")
        if size not in ["M1", "L1"]:
            raise ValueError(f"Invalid size given: {size}")

        duration = round(duration * 10000.0) / 10000.0

        super().__init__(measure, position, NoteType.touch_hold)
        self.duration = duration
        self.is_firework = is_firework
        self.region = region
        self.size = size

    def to_str(self, resolution: int) -> str:
        measure = measure_to_ma2_time(self.measure, resolution)
        template = "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}"
        name = "THO"
        duration = round(self.duration * resolution)
        fireworks = 1 if self.is_firework else 0
        return template.format(
            name,
            measure[0],
            measure[1],
            self.position,
            duration,
            self.region,
            fireworks,
            self.size,
        )


class BPM(Event):
    def __init__(self, measure: float, bpm: float) -> None:
        """Produces a ma2 BPM event.

        Note:
            Please use MaiMa2 class' set_bpm method
            for adding BPM events.

        Args:
            measure: Time when the BPM change happens,
                     in terms of measures.
            bpm: The tempo in beats per minute.

        Raises:
            ValueError: When bpm is not positive.

        Examples:
            A bpm of 220 at measure 3.

            >>> ma2_bpm = BPM(3, 220)
        """
        if bpm <= 0:
            raise ValueError(f"BPM is not positive: {bpm}")

        super().__init__(measure, EventType.bpm)
        self.bpm = bpm

    def to_str(self, resolution: int) -> str:
        if self.measure == 0.0:
            measure = (0, 0)
        else:
            measure = measure_to_ma2_time(self.measure, resolution)

        template = "BPM\t{}\t{}\t{:.3f}"
        return template.format(measure[0], measure[1], self.bpm)


class Meter(Event):
    def __init__(
        self,
        measure: float,
        meter_numerator: int,
        meter_denominator: int,
    ) -> None:
        """Produces a ma2 meter signature event.

        Note:
            Please use MaiMa2 class' set_meter method
            for adding meter signature events.

        Args:
            measure: Time when the meter change happens,
                     in terms of measures.
            meter_numerator: The upper numeral in a meter/time
                             signature.
            meter_denominator: The lower numeral in a meter/time
                             signature.

        Raises:
            ValueError: When meter_numerator or
                        meter_denominator is not positive.
        """
        if meter_numerator <= 0:
            raise ValueError("Meter numerator is not positive: {meter_numerator}")
        if meter_denominator <= 0:
            raise ValueError(f"Meter denominator is not positive {meter_denominator}")

        super().__init__(measure, EventType.meter)
        self.numerator = meter_numerator
        self.denominator = meter_denominator

    def to_str(self, resolution: int) -> str:
        if self.measure == 0.0:
            measure = (0, 0)
        else:
            measure = measure_to_ma2_time(self.measure, resolution)

        template = "MET\t{}\t{}\t{}\t{}"
        return template.format(measure[0], measure[1], self.numerator, self.denominator)


def measure_to_ma2_time(measure: float, resolution: int) -> Tuple[int, int]:
    """Convert measure in decimal form to ma2's format.

    Ma2 uses integer timing for its measures. It does so by taking
    the fractional part and multiplying it to the chart's
    resolution, then rounding it up. For durations like holds
    and slides, both the whole and fractional part are
    multiplied by the charts resolution.

    Args:
        measure: The time a note happened or duration.
        resolution: The number of ticks equal to one measure.

    Returns:
        A tuple (WHOLE_PART, FRACTIONAL_PART).

    Raises:
        ValueError: When measure is negative.

    Examples:
        Convert a note that happens in measure 2.5 in ma2 time
        with resolution of 384.

        >>> print(measure_to_ma2_time(2.5))
        (2, 192)

        Convert a duration of 3.75 measures in ma2 time with
        resolution of 500.

        >>> my_resolution = 500
        >>> (whole, dec) = measure_to_ma2_time(3.75, my_resolution)
        >>> whole*my_resolution + dec
        1875.0
    """
    if measure < 0:
        raise ValueError("Measure is negative. " + str(measure))

    (decimal_part, whole_part) = math.modf(measure)
    decimal_part = round(decimal_part * resolution)

    return int(whole_part), decimal_part


def check_slide(pattern: int, start_position: int, end_position: int):
    """Function that checks a slide if it's valid. Will raise a ValueError if given
    a slide that will crash the game or has undefined behaviour.

    Args:
        pattern: The slide pattern of a slide.
        start_position: The button where a slide begins.
        end_position: The button where a slide ends.

    Raises:
        ValueError: When given a slide that will crash the game or has undefined
            behaviour.
    """
    if not (0 < pattern < 14):
        raise ValueError(f"Invalid slide pattern {pattern}")
    if not (0 <= start_position <= 7):
        raise ValueError(f"Invalid start position {start_position}")
    if not (0 <= end_position <= 7):
        raise ValueError(f"Invalid end position {end_position}")

    distance_cw = slide_distance(start_position, end_position, is_cw=True)
    distance_ccw = slide_distance(start_position, end_position, is_cw=False)

    if pattern == 1:
        if not (distance_cw > 1 and distance_ccw > 1):
            raise ValueError(
                "Distance between start and end position must be greater than 1 in SI_."
            )
    elif pattern in [6, 7, 13]:
        if distance_cw != 4:
            raise ValueError(
                "Start and end position must be opposite of each other in SSL, SSR, or SF_."
            )
    elif pattern == 8:
        if start_position == end_position:
            raise ValueError(
                "Start and end position must not be equal to each other in SV_."
            )
    elif pattern == 11:
        if not 0 < distance_cw < 5:
            raise ValueError("Clockwise distance must be between 0 and 5 in SLL.")
    elif pattern == 12:
        if not 0 < distance_ccw < 5:
            raise ValueError(
                "Counter-clockwise distance must be between 0 and 5 in SLR."
            )



from typing import List

from .ma2note import note_dict, slide_dict

_ignored_v1 = [
    "VERSION",
    "FES_MODE",
    "BPM_DEF",
    "MET_DEF",
    "CLK_DEF",
    "CLK",
    "COMPATIBLE_CODE",
    "T_REC_TAP",
    "T_REC_BRK",
    "T_REC_XTP",
    "T_REC_HLD",
    "T_REC_XHO",
    "T_REC_STR",
    "T_REC_BST",
    "T_REC_XST",
    "T_REC_TTP",
    "T_REC_THO",
    "T_REC_SLD",
    "T_REC_ALL",
    "T_NUM_TAP",
    "T_NUM_BRK",
    "T_NUM_HLD",
    "T_NUM_SLD",
    "T_NUM_ALL",
    "T_JUDGE_TAP",
    "T_JUDGE_HLD",
    "T_JUDGE_SLD",
    "T_JUDGE_ALL",
    "TTM_EACHPAIRS",
    "TTM_SCR_TAP",
    "TTM_SCR_BRK",
    "TTM_SCR_HLD",
    "TTM_SCR_SLD",
    "TTM_SCR_ALL",
    "TTM_SCR_S",
    "TTM_SCR_SS",
    "TTM_RAT_ACV",
]


def parse_v1(ma2, values: List[str]) -> None:
    """Ma2 line parser for versions 1.02.00 and 1.03.00 currently."""
    # For notes and events, the first value is a 3-character text
    line_type = values[0]
    # Create a list of all valid ma2 note and slide types
    if line_type in _ignored_v1:
        # Ignore some parts of the header and all summary statistics lines
        return
    if line_type == "RESOLUTION":
        # Set the max number of ticks in a measure
        ma2._resolution = int(values[1])
    elif line_type == "BPM":
        # Set the BPM for a measure
        measure = float(values[1]) + float(values[2]) / ma2.resolution
        bpm = float(values[3])
        ma2.set_bpm(measure, bpm)
    elif line_type == "MET":
        # Set MET event
        measure = float(values[1]) + float(values[2]) / ma2.resolution
        ma2.set_meter(measure, int(values[3]), int(values[4]))
    elif line_type in list(note_dict.keys()):
        _handle_notes_v1(ma2, values)
    elif line_type in list(slide_dict.keys()):
        _handle_slides_v1(ma2, values)
    else:
        print(f"Warning: Ignoring unknown line type {line_type}")


def _handle_notes_v1(ma2, values: List[str]) -> None:
    line_type = values[0]
    measure = float(values[1]) + float(values[2]) / ma2.resolution
    position = int(values[3])
    if line_type in ["TAP", "BRK", "XTP", "STR", "BST", "XST"]:
        is_break = line_type in ["BRK", "BST"]
        is_ex = line_type in ["XTP", "XST"]
        is_star = line_type in ["STR", "BST", "XST"]
        ma2.add_tap(measure, position, is_break, is_star, is_ex)
    elif line_type in ["XHO", "HLD"]:
        is_ex = line_type == "XHO"
        duration = float(values[4]) / ma2.resolution
        ma2.add_hold(measure, position, duration, is_ex)
    elif line_type == "TTP":
        region = values[4]
        is_firework = values[5] == "1"
        size = values[6] if len(values) > 6 else "M1"
        ma2.add_touch_tap(measure, position, region, is_firework, size)
    elif line_type == "THO":
        duration = float(values[4]) / ma2.resolution
        region = values[5]
        is_firework = values[6] == "1"
        size = values[7] if len(values) > 7 else "M1"
        ma2.add_touch_hold(measure, position, region, duration, is_firework, size)


def _handle_slides_v1(ma2, values: List[str]) -> None:
    line_type = values[0]
    pattern = slide_dict[line_type]
    measure = float(values[1]) + float(values[2]) / ma2.resolution
    start_position = int(values[3])
    delay = int(values[4]) / ma2.resolution
    duration = int(values[5]) / ma2.resolution
    end_position = int(values[6])
    ma2.add_slide(measure, start_position, end_position, duration, pattern, delay)



import os
import gzip
from typing import Union
from binascii import unhexlify
from Crypto.Cipher import AES


def finale_file_decrypt(path: str, key: Union[str, bytes]) -> bytes:
    with open(path, "rb") as f:
        iv = f.read(0x10)
        ciphertext = f.read()

    return finale_decrypt(key=key, iv=iv, ciphertext=ciphertext)


def finale_file_encrypt(path: str, key: Union[str, bytes]) -> bytes:
    with open(path, "rb") as f:
        rawtext = f.read()

    return finale_encrypt(plaintext=rawtext, key=key)


def finale_decrypt(
    key: Union[str, bytes],
    iv: bytes,
    ciphertext: bytes,
) -> bytes:
    if not isinstance(key, bytes):
        key = int(key.replace(" ", ""), 0).to_bytes(0x10, "big")
    if len(key) != 0x10:
        raise ValueError("Invalid key length")

    cipher = AES.new(key, AES.MODE_CBC, iv)
    gzipdata = cipher.decrypt(ciphertext)

    num_padding = gzipdata[-1]
    # Remove padding if there is
    if num_padding > 0:
        gzipdata = gzipdata[:-num_padding]

    # Prefix gzip magic number if it doesn't already
    # gzip.decompress will fail if there's no gzip magic number
    if gzipdata[:2] != b"\x1f\x8b":
        gzipdata = b"\x1f\x8b" + gzipdata

    return gzip.decompress(gzipdata)[0x10:]


def finale_encrypt(
    key: Union[str, bytes],
    plaintext: bytes,
) -> bytes:
    if not isinstance(key, bytes):
        key = int(key.replace(" ", ""), 0).to_bytes(0x10, "big")
    if len(key) != 0x10:
        raise ValueError("Invalid key length")

    JUNK = unhexlify("4b67ca1eebc78fb9964f781019bc4903")
    encoded = JUNK + plaintext

    gzipdata = gzip.compress(encoded)
    if len(gzipdata) % 0x10 != 0:
        amount = 0x10 - (len(gzipdata) % 0x10)
        padding = amount.to_bytes(1, "big") * amount
        gzipdata += padding

    iv = os.urandom(0x10)
    cipher = AES.new(key, AES.MODE_CBC, iv)

    return iv + cipher.encrypt(gzipdata)



import math
from typing import Union, Optional

from .sxtchart import SxtChartType
from ..event import MaiNote, NoteType
from ..tool import slide_distance

srt_note_to_later_note = {
    # 0 (tap notes) can be star, break star, start slide, or tap note
    2: 2,
    4: 3,
    128: 128,
}

srt_szt_template = "{:.6f}, {:.6f}, {:.6f}, {:2d}, {:3d}, {:3d}, {:2d},\n"
sct_template = "{:.4f}, {:.4f}, {:.4f}, {:2d}, {:3d}, {:3d}, {:2d}, {:2d},\n"
sdt_template = "{:.4f}, {:.4f}, {:.4f}, {:2d}, {:3d}, {:3d}, {:2d}, {:2d}, {:.4f},\n"


class TapNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        is_break: bool = False,
        is_star: bool = False,
    ) -> None:
        """Produces an sxt tap note.

        Note:
            Please use MaiSxt class' add_tap method for adding
            tap notes.

        Args:
            measure: Time when the tap happens, in terms of measures.
            position: Button where the tap note happens.
            is_break: Whether a tap note is a break note.
        """
        if is_break and is_star:
            super().__init__(measure, position, NoteType.break_star)
            self.amount = 0
        elif is_break:
            super().__init__(measure, position, NoteType.break_tap)
        elif is_star:
            super().__init__(measure, position, NoteType.star)
            self.amount = 0
        else:
            super().__init__(measure, position, NoteType.tap)

    def __str__(self) -> str:
        return sdt_note_to_str(self)
        # return self.to_str(SxtChartType.SDT)

    def to_str(
        self,
        chart_type: SxtChartType,
        srt_slide_duration: Optional[float] = None,
        srt_slide_id: Optional[int] = None,
        srt_slide_pattern: Optional[int] = None,
    ) -> str:
        raise NotImplementedError

        measure = math.modf(self.measure)
        is_star = self.note_type in [NoteType.break_star, NoteType.star]

        if chart_type is SxtChartType.SRT:
            if self.note_type is NoteType.break_star:
                print(
                    "Warning: Converting break star to regular star. Report an issue if SRT supports break stars."
                )

            if is_star and (
                srt_slide_duration is None
                or srt_slide_id is None
                or srt_slide_pattern is None
            ):
                raise ValueError(
                    "Slide duration, id, or pattern is not given for SRT star note."
                )

            return srt_szt_template.format(
                measure[1],
                measure[0],
                0.0 if not is_star else srt_slide_duration,
                self.position,
                4 if self.note_type is NoteType.break_tap else 0,
                0 if srt_slide_id is None else srt_slide_id,
                0 if srt_slide_pattern is None else srt_slide_pattern,
            )
        elif chart_type is SxtChartType.SZT:
            return srt_szt_template.format(
                measure[1],
                measure[0],
                0.0625,
                self.position,
                self.note_type.value,
                0,
                0,
            )
        elif chart_type is SxtChartType.SCT:
            return sct_template.format(
                measure[1],
                measure[0],
                0.0625,
                self.position,
                self.note_type.value,
                0,
                0,
                0 if not is_star else self.amount,
            )
        else:
            return sdt_template.format(
                measure[1],
                measure[0],
                0.0625,
                self.position,
                self.note_type.value,
                0,
                0,
                0 if not is_star else self.amount,
                0.0,
            )


class SlideStartNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        pattern: int,
        duration: float,
        slide_id: int,
        delay: float = 0.25,
    ) -> None:
        """Produces an sxt slide start note.

        Note:
            Please use MaiSxt class' add_slide method for adding
            slide start and slide end.

        Args:
            measure: Time when the slide starts, in terms of measures.
            position: Button where the slide starts.
            slide_id: Unique non-zero integer for slide start
                and end pair.
            pattern: Numerical representation of the slide pattern.
            duration: Total duration of the slide, in terms of
                measures. Includes slide delay.
            delay: Duration from when the slide appears and when it
                starts to move, in terms of measures.
        """
        if slide_id <= 0:
            raise ValueError("Slide id is not positive " + str(slide_id))
        if pattern <= 0:
            raise ValueError("Slide pattern is not positive " + str(pattern))
        if duration <= 0:
            raise ValueError("Slide duration is not positive " + str(duration))
        if delay < 0:
            raise ValueError("Slide delay is negative " + str(delay))

        duration = round(10000.0 * duration) / 10000.0
        delay = round(10000.0 * delay) / 10000.0
        super().__init__(measure, position, NoteType.start_slide)
        self.slide_id = slide_id
        self.pattern = pattern
        self.delay = delay
        self.duration = duration

    def __str__(self) -> str:
        return sdt_note_to_str(self)

    # def to_str(self, slide_delay: float):


class SlideEndNote(MaiNote):
    def __init__(
        self, measure: float, position: int, pattern: int, slide_id: int
    ) -> None:
        """Produces an sxt slide end note.

        Note:
            Please use MaiSxt class' add_slide method for adding
            slide start and slide end.

        Args:
            measure: Time when the slide ends, in terms of measures.
            position: Button where the slide ends.
            slide_id: Unique non-zero integer for slide start
                and end pair.
            pattern: Numerical representation of the slide pattern.
        """
        if slide_id <= 0:
            raise ValueError("Slide id is not positive " + str(slide_id))
        if pattern <= 0:
            raise ValueError("Slide pattern is not positive " + str(pattern))

        measure = round(10000 * measure) / 10000
        super().__init__(measure, position, NoteType.end_slide)
        self.slide_id = slide_id
        self.pattern = pattern

    def __str__(self) -> str:
        return sdt_note_to_str(self)


class HoldNote(MaiNote):
    def __init__(self, measure: float, position: int, duration: float) -> None:
        """Produces an sxt hold note.

        Note:
            Please use MaiSxt class' add_hold method for adding
            hold notes.

        Args:
            measure: Time when the hold starts, in terms of measures.
            position: Button where the hold happens.
            duration: Total duration of the hold, in terms of measures.
        """
        if duration < 0:
            raise ValueError(f"Hold duration is negative: {duration}")

        measure = round(10000.0 * measure) / 10000.0
        duration = round(10000.0 * duration) / 10000.0
        super().__init__(measure, position, NoteType.hold)
        self.duration = duration

    def __str__(self) -> str:
        return sdt_note_to_str(self)


def sdt_note_to_str(
    note: Union[TapNote, HoldNote, SlideEndNote, SlideStartNote]
) -> str:
    """Prints note into sxt-compatible lines.

    Args:
        note: A MaiNote to be converted to a sxt string.

    Returns:
        A single line string.
    """
    measure = math.modf(note.measure)
    note_type = note.note_type
    if isinstance(note, (HoldNote, SlideStartNote)):
        note_duration = note.duration
    elif isinstance(note, SlideEndNote):
        note_duration = 0
    else:
        note_duration = 0.0625

    if isinstance(note, (SlideStartNote, SlideEndNote)):
        slide_id = note.slide_id
        pattern = note.pattern
    else:
        slide_id = 0
        pattern = 0

    if isinstance(note, TapNote) and note.note_type in [
        NoteType.star,
        NoteType.break_star,
    ]:
        slide_amount = note.amount
    else:
        slide_amount = 0

    delay = 0.0 if not isinstance(note, SlideStartNote) else note.delay
    line_template = "{:.4f}, {:.4f}, {:.4f}, {:2d}, {:3d}, {:3d}, {:2d}, {:2d}, {:.4f},"
    result = line_template.format(
        measure[1],
        measure[0],
        note_duration,
        note.position,
        note_type.value,
        slide_id,
        pattern,
        slide_amount,
        delay,
    )
    return result


def check_slide(
    pattern: int,
    start_position: int,
    end_position: int,
    chart_type: SxtChartType = SxtChartType.SDT,
):
    """Function that checks a slide if it's valid. Will raise a ValueError if given
    a slide that will crash the game or has undefined behaviour.

    Args:
        pattern: The slide pattern of a slide. Should be ints corresponding to slides from SZT and later.
        start_position: The button where a slide begins.
        end_position: The button where a slide ends.
        chart_type: Optional chart type. Used only for checking SRT slides.

    Raises:
        ValueError: When given a slide that will crash the game or has undefined
            behaviour.
    """
    if not (0 <= start_position <= 7):
        raise ValueError(f"Invalid start position {start_position}")
    if not (0 <= end_position <= 7):
        raise ValueError(f"Invalid end position {end_position}")
    if chart_type is SxtChartType.SRT and not (1 <= pattern <= 3):
        raise ValueError(f"Invalid pattern for SRT chart {pattern}")
    if chart_type is not SxtChartType.SRT and not (1 <= pattern <= 13):
        raise ValueError(f"Invalid pattern for non-SRT chart {pattern}")

    # Helper variables
    distance_cw = slide_distance(start_position, end_position, is_cw=True)
    distance_ccw = slide_distance(start_position, end_position, is_cw=False)

    if pattern == 1:
        # Straight slide's end position should at least be two places away
        if not 2 <= distance_cw <= 6:
            raise ValueError(
                "Distance between start and end position should be greater than 1 in pattern 1"
            )
    elif pattern == 2 and chart_type is SxtChartType.SRT:
        # CCW around the judgement ring in SRT can only do 3 places max
        if not distance_ccw <= 3:
            raise ValueError("SRT can only do distances of 3 places max in pattern 2")
    elif pattern == 3 and chart_type is SxtChartType.SRT:
        # CW around the judgement ring in SRT can only do 3 places max
        if not distance_cw <= 3:
            raise ValueError("SRT can only do distances of 3 places max in pattern 3")
    elif pattern in [6, 7] and distance_cw != 4:
        # Zigzags end_position should be opposite of start_position
        raise ValueError(
            "End position is not opposite of start position in pattern 6 or 7"
        )
    elif pattern == 11:
        if not distance_ccw >= 4:
            raise ValueError(f"CCW distance is less than 4 in pattern 11")
    elif pattern == 12:
        if not distance_cw >= 4:
            raise ValueError(f"CW distance is less than 4 in pattern 12")
    elif pattern == 13 and distance_cw != 4:
        raise ValueError("End position is not opposite of start position in pattern 13")



from __future__ import annotations

import math
import re
from typing import Union, List, Dict

from .sxtnote import (
    TapNote,
    HoldNote,
    SlideStartNote,
    SlideEndNote,
    check_slide,
)
from ..event import NoteType
from ..tool import measure_to_second, second_to_measure, offset_arg_to_measure


class MaiSxt:
    """A class that represents an sxt (and predecessors) chart.
    Only contains notes, and does not include information
    such as song name, chart difficulty, composer, chart maker, etc.
    It only contains enough information to build a working sxt
    chart file.

    Attributes:
        bpm: Singular BPM in which the chart is written in.
        notes: Contains notes of the chart.
    """

    def __init__(self, bpm: float) -> None:
        if bpm <= 0:
            raise ValueError(f"BPM is not positive {bpm}")

        self.bpm = bpm
        self.notes: List[Union[TapNote, HoldNote, SlideStartNote, SlideEndNote]] = []
        self.start_slide_notes: Dict[int, Dict[str, Union[int, float]]] = {}
        self.slide_count = 1

    @classmethod
    def open(cls, path: str, bpm: float, encoding: str = "utf-8") -> MaiSxt:
        sdt = cls(bpm=bpm)
        with open(path, "r", encoding=encoding) as file:
            for line in file:
                if line in ["\n", "\r\n"]:
                    continue

                if re.search(r"\.srt", path) is None:
                    sdt.parse_line(line)
                else:
                    sdt.parse_srt_line(line)

        return sdt

    def parse_line(self, line: str) -> MaiSxt:
        """Parse a non-SRT comma-separated line.

        Args:
            line: A non-SRT comma-separated line.

        Raises:
            ValueError: When the number of columns are not between 7 and 9.
                When an end slide is declared before an beginning slide.
                When an unknown note type is given.
        """
        values = line.rstrip().rstrip(",").replace(" ", "").split(",")
        if not (7 <= len(values) <= 9):
            raise ValueError(f"Line has invalid number of columns {len(values)}")

        measure = float(values[0]) + float(values[1])
        position = int(values[3])
        note_type = int(values[4])

        if note_type in [1, 3, 4, 5]:
            # Regular tap note, break tap note, star note, break star note
            is_star = note_type in [4, 5]
            is_break = note_type in [3, 5]
            self.add_tap(
                measure=measure, position=position, is_break=is_break, is_star=is_star
            )
        elif note_type == 2:
            # Hold note
            duration = float(values[2])
            self.add_hold(measure, position, duration)
        elif note_type == 0:  # Start slide
            if len(values) == 9:
                # SDT includes delay
                delay = float(values[8])
            else:
                delay = 0.25

            duration = float(values[2])
            slide_id = int(values[5])
            pattern = int(values[6])
            start_slide = {
                "position": position,
                "measure": measure,
                "duration": duration,
                "delay": delay,
                "pattern": pattern,
            }
            self.start_slide_notes[slide_id] = start_slide
        elif note_type == 128:  # End slide
            slide_id = int(values[5])

            # Get information about corresponding start slide
            if slide_id not in self.start_slide_notes:
                raise ValueError("End slide is declared before slide start!")

            start_slide = self.start_slide_notes[slide_id]
            start_measure = start_slide["measure"]
            start_position = start_slide["position"]
            duration = start_slide["duration"]
            delay = start_slide["delay"]
            slide_pattern = start_slide["pattern"]
            self.add_slide(
                start_measure,
                int(start_position),
                position,
                duration,
                int(slide_pattern),
                delay,
            )
        else:
            raise ValueError("Unknown note type {}".format(note_type))

        return self

    def parse_srt_line(self, line: str) -> MaiSxt:
        """Parse an SRT comma-separated line.

        Args:
            line: An SRT comma-separated line.

        Raises:
            ValueError: When the number of columns are not between 7 and 9.
                When an end slide is declared before an beginning slide.
                When an unknown note type is given.
        """
        srt_slide_to_later_dict = {
            0: 1,
            1: 3,
            2: 2,
        }
        values = line.rstrip().rstrip(",").replace(" ", "").split(",")
        if len(values) != 7:
            raise ValueError(f"SRT should have 7 columns. Given: {len(values)}")

        measure = float(values[0]) + float(values[1])
        position = int(values[3])
        note_type = int(values[4])

        if note_type in [0, 4]:
            # 0: Regular tap note or star and start slide note
            # 4: Break tap note
            slide_id = int(values[5])
            if slide_id != 0:
                # Tap notes with a non-zero slide id are stars and start slide
                self.add_tap(measure=measure, position=position, is_star=True)
                duration = float(values[2])
                # Slide patterns in SZT, and later, starts at 1
                pattern = srt_slide_to_later_dict[int(values[6])]
                start_slide: Dict[str, Union[float, int]] = {
                    "position": position,
                    "measure": measure,
                    "duration": duration,
                    "pattern": pattern,
                }
                self.start_slide_notes[slide_id] = start_slide
            else:
                is_break = note_type == 4
                self.add_tap(measure=measure, position=position, is_break=is_break)
        elif note_type == 2:  # Hold note
            duration = float(values[2])
            self.add_hold(measure, position, duration)
        elif note_type == 128:
            slide_id = int(values[5])
            # Get information about corresponding start slide
            if slide_id not in self.start_slide_notes:
                raise RuntimeError("End slide is declared before slide start!")

            start: Dict[str, Union[float, int]] = self.start_slide_notes[slide_id]
            start_measure = start["measure"]
            start_position = start["position"]
            duration = start["duration"]
            slide_pattern = start["pattern"]
            self.add_slide(
                start_measure,
                int(start_position),
                position,
                duration,
                int(slide_pattern),
            )
        else:
            raise ValueError(f"Unknown note type {note_type}")

        return self

    def add_tap(
        self,
        measure: float,
        position: int,
        is_break: bool = False,
        is_star: bool = False,
    ) -> MaiSxt:
        """Adds a tap note to the list of notes.

        Args:
            measure: Time when the tap happens, in terms of measures.
            position: Button where the tap note happens.
            is_break: Whether a tap note is a break note.
            is_star: Whether a tap note is a star note.

        Examples:
            Add a regular tap note at measure 1 at button 2,
            and a break tap note at measure 2 at button 7.

            >>> sxt = MaiSxt()
            >>> sxt.add_tap(1, 2)
            >>> sxt.add_tap(2, 7, is_break=True)
        """
        tap_note = TapNote(
            measure=measure, position=position, is_break=is_break, is_star=is_star
        )
        self.notes.append(tap_note)

        return self

    def del_tap(self, measure: float, position: int) -> MaiSxt:
        """Deletes a tap note from the list of notes.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the tap note happens.

        Examples:
            Create a break tap note at measure 26.75 at button 4. Then delete it.

            >>> sxt = MaiSxt()
            >>> sxt.add_tap(26.75, 4, is_break=True)
            >>> sxt.del_tap(26.75, 4)
        """
        tap_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == position
        ]
        for note in tap_notes:
            self.notes.remove(note)

        return self

    def add_hold(
        self,
        measure: float,
        position: int,
        duration: float,
    ) -> MaiSxt:
        """Adds a hold note to the list of notes.

        Args:
            measure: Time when the hold starts, in terms of measures.
            position: Button where the hold happens.
            duration: Total duration of the hold, in terms of measures.

        Examples:
            Add a regular hold note at button 5 at measure 1.5, with
            duration of 2.75 measures.

            >>> sxt = MaiSxt()
            >>> sxt.add_hold(1.5, 5, 2.75)
        """
        hold_note = HoldNote(measure=measure, position=position, duration=duration)
        self.notes.append(hold_note)

        return self

    def del_hold(
        self,
        measure: float,
        position: int,
    ) -> MaiSxt:
        """Deletes the matching hold note in the list of notes. If there are multiple
        matches, all matching notes are deleted. If there are no match, nothing happens.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the hold note happens.

        Examples:
            Add a regular hold note at button 0 at measure 3.25 with duration of 2 measures
            and delete it.

            >>> sxt = MaiSxt()
            >>> sxt.add_hold(3.25, 0, 2)
            >>> sxt.del_hold(3.25, 0)
        """
        hold_notes = [
            x
            for x in self.notes
            if isinstance(x, HoldNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == position
        ]
        for note in hold_notes:
            self.notes.remove(note)

        return self

    def add_slide(
        self,
        measure: float,
        start_position: int,
        end_position: int,
        duration: float,
        pattern: int,
        delay: float = 0.25,
        slide_check: bool = True,
    ) -> MaiSxt:
        """Adds both a start slide and end slide note to the list of notes.
        Note: SXT slide duration includes delay and should not be less than its delay.

        Args:
            measure: Time when the slide starts, in
                terms of measures.
            start_position: Button where the slide starts.
            end_position: Button where the slide ends.
            duration: Total duration of the slide, in terms of
                measures. Includes slide delay.
            pattern: Numerical representation of the slide pattern.
            delay: Duration from when the slide appears and when it
                starts to move, in terms of measures. Defaults to 0.25.
            slide_check: When set to true, will check validity of slides.

        Raises:
            ValueError: When a slide is invalid and crashes the game (Can be disabled by setting check_slide to False.)
                When fail_on_undefined is True and given a slide with undefined behaviour.

        Examples:
            Add a slide at measure 2 from button 6 to button 3 with
            duration of 1.75 measures, delay of 0.25 measures,
            pattern of 1.

            >>> sxt = MaiSxt()
            >>> sxt.add_slide(2, 6, 3, 1.75, 1)
        """
        if slide_check:
            check_slide(pattern, start_position, end_position)

        slide_id = self.slide_count
        start_slide = SlideStartNote(
            measure=measure,
            position=start_position,
            pattern=pattern,
            duration=duration,
            slide_id=slide_id,
            delay=delay,
        )
        end_measure = measure + duration
        end_slide = SlideEndNote(
            measure=end_measure,
            position=end_position,
            pattern=pattern,
            slide_id=slide_id,
        )
        self.notes.append(start_slide)
        self.notes.append(end_slide)
        self.slide_count += 1

        star_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and x.note_type in [NoteType.star, NoteType.break_star]
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == start_position
        ]
        for star_note in star_notes:
            star_note.amount += 1

        return self

    def del_slide(
        self,
        measure: float,
        start_position: int,
        end_position: int,
    ) -> MaiSxt:
        start_slides = [
            x
            for x in self.notes
            if isinstance(x, SlideStartNote)
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == start_position
        ]
        end_slides: List[SlideEndNote] = []
        for note in start_slides:
            slides = [
                x
                for x in self.notes
                if isinstance(x, SlideEndNote)
                and x.slide_id == note.slide_id
                and x.position == end_position
            ]
            end_slides += slides

        correct_start_slides = []
        for note in end_slides:
            slides = [
                x
                for x in self.notes
                if isinstance(x, SlideStartNote) and x.slide_id == note.slide_id
            ]
            correct_start_slides += slides

        star_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and x.note_type in [NoteType.star, NoteType.break_star]
            and math.isclose(x.measure, measure, abs_tol=0.0001)
            and x.position == start_position
        ]

        for note in correct_start_slides:
            self.notes.remove(note)

        for note in end_slides:
            self.notes.remove(note)

        for star_note in star_notes:
            star_note.amount -= 1

        return self

    def offset(self, offset: Union[float, str]) -> MaiSxt:
        offset = offset_arg_to_measure(offset, self.second_to_measure)

        for note in self.notes:
            note.measure = round((note.measure + offset) * 10000.0) / 10000.0

        return self

    def measure_to_second(self, measure: float) -> float:
        return measure_to_second(measure, [(0.0, self.bpm)])

    def second_to_measure(self, seconds: float) -> float:
        return second_to_measure(seconds, [(0.0, self.bpm)])

    def export(self) -> str:
        """Generates an sxt text from all the notes defined.

        Returns:
            A multiline string. The returned
            string is a complete and functioning sxt text and should
            be stored as-is in a text file with an sxt file extension.
        """
        self.notes.sort()
        return "\n".join([str(note) for note in self.notes]) + "\n"



import os
import argparse
import re
import traceback
import sys

import maiconverter
from maiconverter.maicrypt import finale_file_encrypt, finale_file_decrypt
from maiconverter.maima2 import MaiMa2
from maiconverter.maisxt import MaiSxt
from maiconverter.simai import parse_file, SimaiChart
from maiconverter.converter import (
    ma2_to_sdt,
    ma2_to_simai,
    sdt_to_ma2,
    sdt_to_simai,
    simai_to_ma2,
    simai_to_sdt,
)


def file_path(string):
    if os.path.exists(string):
        return string.rstrip("/\\")

    raise FileNotFoundError(string)


# Only accepts directory paths
def dir_path(string):
    if os.path.isdir(string):
        return string.rstrip("/\\")

    raise NotADirectoryError(string)


def crypto(args, output):
    if args.key is None:
        raise RuntimeError("Key not supplied")

    if os.path.isdir(args.path):
        files = os.listdir(args.path)
        files = [
            os.path.join(args.path, file)
            for file in files
            if os.path.isfile(os.path.join(args.path, file))
        ]

        if args.command == "encrypt":
            if args.database:
                files = [
                    file for file in files if not re.search(r"\.tbl", file) is None
                ]
            else:
                # Only accept ".sdt" ".sct" ".szt" ".srt" files
                files = [
                    file for file in files if not re.search(r"\.s.t", file) is None
                ]
        else:  # decrypt
            if args.database:
                files = [
                    file for file in files if not re.search(r"\.bin", file) is None
                ]
            else:
                # Only accept ".sdb" ".scb" ".szb" ".srb" files
                files = [
                    file for file in files if not re.search(r"\.s.b", file) is None
                ]
    else:
        files = [args.path]

    for file in files:
        if args.database:
            handle_db(file, output, args.command, args.key)
        else:
            handle_file(file, output, args.command, args.key)


def chart_convert(args, output):
    if args.command in ["ma2tosdt", "ma2tosimai"]:
        file_regex = r"\.ma2"
    elif args.command in ["sdttoma2", "sdttosimai"]:
        if args.bpm is None:
            raise RuntimeError("BPM required for SDT file")

        file_regex = r"\.s.t"
    else:
        file_regex = r"\.txt"

    if os.path.isdir(args.path):
        files = os.listdir(args.path)
        files = [
            os.path.join(args.path, file)
            for file in files
            if os.path.isfile(os.path.join(args.path, file))
            and re.search(file_regex, file) is not None
        ]
    else:
        files = [args.path]

    for file in files:
        name = os.path.splitext(os.path.basename(file))[0]

        try:
            if args.command in ["ma2tosdt", "ma2tosimai"]:
                handle_ma2(file, name, output, args)
            elif args.command in ["sdttoma2", "sdttosimai"]:
                handle_sxt(file, name, output, args)
            elif args.command in ["simaifiletoma2", "simaifiletosdt"]:
                handle_simai_file(file, output, args)
            else:
                handle_simai_chart(file, name, output, args)

        except:
            print(f"Error occurred processing {file}.")
            raise


def handle_ma2(file, name, output_path, args):
    ma2 = MaiMa2.open(file, encoding=args.encoding)
    if len(args.delay) != 0:
        ma2.offset(args.delay)

    if args.command == "ma2tosdt":
        output = ma2_to_sdt(ma2, convert_touch=args.convert_touch)
        ext = ".sdt"
    else:
        output = ma2_to_simai(ma2)
        ext = ".txt"

    with open(
        os.path.join(output_path, name + ext), "w+", newline="\r\n", encoding="utf-8"
    ) as out:
        if isinstance(output, SimaiChart):
            out.write(output.export(max_den=args.max_divisor))
        else:
            out.write(output.export())


def handle_sxt(file, name, output_path, args):
    sxt = MaiSxt.open(file, encoding=args.encoding, bpm=args.bpm)
    if len(args.delay) != 0:
        sxt.offset(args.delay)

    if args.command == "sdttoma2":
        output = sdt_to_ma2(sxt)
        ext = ".ma2"
    else:
        output = sdt_to_simai(sxt)
        ext = ".txt"

    with open(
        os.path.join(output_path, name + ext), "w+", newline="\r\n", encoding="utf-8"
    ) as out:
        if isinstance(output, SimaiChart):
            out.write(output.export(max_den=args.max_divisor))
        else:
            out.write(output.export(resolution=args.resolution))


def handle_simai_chart(file, name, output_path, args):
    with open(file, "r", encoding=args.encoding) as f:
        chart_text = f.read()

    simai = SimaiChart.from_str(chart_text, message=f"Parsing Simai chart at {file}...")
    if len(args.delay) != 0:
        simai.offset(args.delay)

    if args.command == "simaitosdt":
        ext = ".sdt"
        converted = simai_to_sdt(simai, convert_touch=args.convert_touch)
    else:
        ext = ".ma2"
        converted = simai_to_ma2(simai)

    with open(
        os.path.join(output_path, name + ext), "w+", newline="\r\n", encoding="utf-8"
    ) as out:
        if isinstance(converted, MaiSxt):
            out.write(converted.export())
        else:
            out.write(converted.export(resolution=args.resolution))


def handle_simai_file(file, output_path, args):
    title, charts = parse_file(file, encoding=args.encoding)
    for i, chart in enumerate(charts):
        diff, simai_chart = chart
        if len(args.delay) != 0:
            simai_chart.offset(args.delay)

        try:
            if args.command == "simaifiletosdt":
                ext = ".sdt"
                converted = simai_to_sdt(simai_chart, convert_touch=args.convert_touch)
            else:
                ext = ".ma2"
                converted = simai_to_ma2(simai_chart)

            name = title + f"_{diff}"
            with open(
                os.path.join(output_path, name + ext),
                "w+",
                newline="\r\n",
                encoding="utf-8",
            ) as out:
                if isinstance(converted, MaiSxt):
                    out.write(converted.export())
                else:
                    out.write(converted.export(resolution=args.resolution))
        except:
            print(f"Error processing {i + 1} chart of file.")
            raise


def handle_file(input_path, output_dir, command, key):
    file_name = os.path.splitext(os.path.basename(input_path))[0]
    if command == "encrypt":
        file_ext = os.path.splitext(input_path)[1].replace("t", "b")
        output = finale_file_encrypt(input_path, key)

    else:
        file_ext = os.path.splitext(input_path)[1].replace("b", "t")
        output = finale_file_decrypt(input_path, key)

    with open(os.path.join(output_dir, file_name + file_ext), "wb") as f:
        f.write(output)


def handle_db(input_path, output_dir, command, key):
    file_name = os.path.splitext(os.path.basename(input_path))[0]

    if command == "encrypt":
        file_ext = ".bin"
        output = finale_file_encrypt(key, input_path)
    else:
        file_ext = ".tbl"
        output = finale_file_decrypt(key, input_path)

    with open(os.path.join(output_dir, file_name + file_ext), "wb") as f:
        f.write(output)


COMMANDS = [
    "encrypt",
    "decrypt",
    "ma2tosdt",
    "ma2tosimai",
    "sdttoma2",
    "sdttosimai",
    "simaifiletoma2",
    "simaifiletosdt",
    "simaitoma2",
    "simaitosdt",
]


def parse_arg():
    parser = argparse.ArgumentParser(
        description="Tool for converting MaiMai chart formats",
        allow_abbrev=False,
    )
    parser.add_argument(
        "command",
        metavar="command",
        type=str,
        choices=COMMANDS,
        help="Specify whether to encrypt or decrypt",
    )
    parser.add_argument("path", metavar="input", type=file_path, help="")
    parser.add_argument(
        "-k",
        "--key",
        type=str,
        help="16 byte AES key for encrypt/decrypt (Prepend hex value with 0x)",
    )
    parser.add_argument(
        "--database",
        "-db",
        action="store_const",
        default=False,
        const=True,
        help="Database/s toggle for encrypt/decrypt",
    )
    parser.add_argument(
        "-b",
        "--bpm",
        metavar="Song BPM",
        type=float,
        help="Song BPM for sdttoma2/sdttosimai",
    )
    parser.add_argument(
        "-d",
        "--delay",
        metavar="Delay",
        type=str,
        default="",
        help="Offset a chart by set measures (can be negative)",
    )
    parser.add_argument(
        "-ct",
        "--convert_touch",
        action="store_const",
        default=False,
        const=True,
        help="Toggle to convert touch notes for chart conversion to SDT",
    )
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        default=384,
        help="Output chart resolution for ma2 charts",
    )
    parser.add_argument(
        "-md",
        "--max_divisor",
        type=int,
        default=1000,
        help="Max divisor used in Simai export",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="Output directory",
        type=dir_path,
        help="Path to output. Defaults to /path/to/input/output",
    )
    parser.add_argument(
        "-e",
        "--encoding",
        type=str,
        default="utf-8",
        help="Specify encoding of source file. Defaults to utf-8",
    )

    return parser.parse_args()


def main():
    args = parse_arg()
    print(f"MaiConverter {maiconverter.__version__} by donmai")

    if args.output is None:
        if os.path.isdir(args.path):
            output_dir = os.path.join(args.path, "output")
        else:
            output_dir = os.path.join(os.path.dirname(args.path), "output")
    else:
        output_dir = args.output

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    elif os.path.exists(output_dir) and not os.path.isdir(output_dir):
        raise NotADirectoryError(output_dir)

    try:
        if args.command in ["encrypt", "decrypt"]:
            crypto(args, output_dir)
        else:
            chart_convert(args, output_dir)
    finally:
        print(
            "Finished. Join MaiMai Tea Discord server for more info and tools about MaiMai modding!"
        )
        print("https://discord.gg/WxEMM9dnwR")



from .maima2tomaisxt import ma2_to_sdt
from .maima2tosimai import ma2_to_simai
from .maisxttomaima2 import sdt_to_ma2
from .maisxttosimai import sdt_to_simai
from .simaitomaima2 import simai_to_ma2
from .simaitomaisxt import simai_to_sdt



import argparse
import copy
import os.path

from maiconverter.maisxt import MaiSxt, HoldNote, SlideStartNote


def main():
    parser = argparse.ArgumentParser("sxt to ma2 with bpm changes")
    parser.add_argument("input", type=str)
    parser.add_argument("original_bpm", type=float)
    parser.add_argument("new_bpm", type=float)

    args = parser.parse_args()
    sdt = MaiSxt.open(args.input, bpm=args.new_bpm)
    sdt.notes.sort(key=lambda x: x.measure)

    scale = args.new_bpm / args.original_bpm

    new_notes = []
    previous_measure = 0.0
    equivalent_current_measure = 0.0
    for note in sdt.notes:
        current_measure = note.measure
        equivalent_current_measure += scale * (current_measure - previous_measure)
        previous_measure = current_measure

        note = copy.deepcopy(note)
        note.measure = equivalent_current_measure

        if isinstance(note, HoldNote):
            note.duration *= scale
        elif isinstance(note, SlideStartNote):
            note.duration *= scale
            note.delay *= scale

        new_notes.append(note)

    new_notes.sort()
    sdt.notes = new_notes

    filename, _ = os.path.splitext(args.input)
    with open(filename + f"_bpm{args.new_bpm}.sxt", "w", newline="\r\n") as out:
        out.write(sdt.export())


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import List, Union

from ..maima2 import MaiMa2
from ..maisxt import MaiSxt, TapNote, HoldNote, SlideStartNote, SlideEndNote
from ..event import NoteType


@dataclass
class StartSlide:
    measure: float
    position: int
    duration: float
    delay: float
    slide_id: int


def sdt_to_ma2(
    sdt: MaiSxt,
    fes_mode: bool = False,
) -> MaiMa2:
    ma2 = MaiMa2(fes_mode=fes_mode)
    ma2.set_bpm(0.0, sdt.bpm)
    ma2.set_meter(0.0, 4, 4)
    convert_notes(ma2, sdt.notes)
    return ma2


def convert_notes(
    ma2: MaiMa2,
    sdt_notes: List[Union[TapNote, HoldNote, SlideStartNote, SlideEndNote]],
) -> None:
    start_slides = []
    for sdt_note in sdt_notes:
        note_type = sdt_note.note_type
        if isinstance(sdt_note, TapNote):
            is_break = note_type in [NoteType.break_tap, NoteType.break_star]
            is_star = note_type in [NoteType.star, NoteType.break_star]
            ma2.add_tap(
                measure=sdt_note.measure,
                position=sdt_note.position,
                is_break=is_break,
                is_star=is_star,
            )
        elif isinstance(sdt_note, HoldNote):
            ma2.add_hold(
                measure=sdt_note.measure,
                position=sdt_note.position,
                duration=sdt_note.duration,
            )
        elif isinstance(sdt_note, SlideStartNote):
            # ma2 slide durations does not include the delay unlike in sxt
            start_slide = StartSlide(
                measure=sdt_note.measure,
                position=sdt_note.position,
                duration=sdt_note.duration - sdt_note.delay,
                delay=sdt_note.delay,
                slide_id=sdt_note.slide_id,
            )
            start_slides.append(start_slide)
        elif isinstance(sdt_note, SlideEndNote):
            starts = [
                slide for slide in start_slides if slide.slide_id == sdt_note.slide_id
            ]
            if len(starts) == 0:
                raise Exception("No corresponding start slide")
            if len(starts) > 1:
                raise Exception("Multiple start slides with same slide id")

            start_slide = starts[0]
            ma2.add_slide(
                measure=start_slide.measure,
                start_position=start_slide.position,
                end_position=sdt_note.position,
                duration=start_slide.duration,
                pattern=sdt_note.pattern,
                delay=start_slide.delay,
            )
        else:
            print("Warning: Unknown note type {}".format(note_type))



from typing import Optional, Tuple

from ..event import Event, EventType, SimaiNote, NoteType
from ..tool import slide_distance, slide_is_cw


# For straightforward slide pattern conversion from simai to sxt/ma2.
# Use simai_pattern_to_int to cover all simai slide patterns.
slide_dict = {
    "-": 1,
    "p": 4,
    "q": 5,
    "s": 6,
    "z": 7,
    "v": 8,
    "pp": 9,
    "qq": 10,
    "w": 13,
}

slide_patterns = [
    "-",
    "^",
    ">",
    "<",
    "p",
    "q",
    "s",
    "z",
    "v",
    "pp",
    "qq",
    "V",
    "w",
]


class TapNote(SimaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        is_break: bool = False,
        is_star: bool = False,
        is_ex: bool = False,
    ) -> None:
        measure = round(100000.0 * measure) / 100000.0
        if is_ex and is_star:
            super().__init__(measure, position, NoteType.ex_star)
        elif is_ex and not is_star:
            super().__init__(measure, position, NoteType.ex_tap)
        elif is_star and is_break:
            super().__init__(measure, position, NoteType.break_star)
        elif is_star and not is_break:
            super().__init__(measure, position, NoteType.star)
        elif is_break:
            super().__init__(measure, position, NoteType.break_tap)
        else:
            super().__init__(measure, position, NoteType.tap)


class HoldNote(SimaiNote):
    def __init__(
        self, measure: float, position: int, duration: float, is_ex: bool = False
    ) -> None:
        if duration < 0:
            raise ValueError(f"Hold duration is negative: {duration}")

        measure = round(100000.0 * measure) / 100000.0
        duration = round(100000.0 * duration) / 100000.0
        if is_ex:
            super().__init__(measure, position, NoteType.ex_hold)
        else:
            super().__init__(measure, position, NoteType.hold)

        self.duration = duration


class SlideNote(SimaiNote):
    def __init__(
        self,
        measure: float,
        start_position: int,
        end_position: int,
        duration: float,
        pattern: str,
        delay: float = 0.25,
        reflect_position: Optional[int] = None,
    ) -> None:
        """Produces a simai slide note.

        Note: Simai slide durations does not include slide delay.

        Args:
            measure (float): Measure where the slide note begins.
            start_position (int): Button where slide begins. [0, 7]
            end_position (int): Button where slide ends. [0, 7]
            duration (float): Total duration, in measures, of
                the slide, including delay.
            pattern (str): Simai slide pattern. If 'V', then
                reflect_position should not be None.
            delay (float, optional): Time duration, in measures,
                from where slide appears and when it starts to move.
                Defaults to 0.25.
            reflect_position (Optional[int], optional): For 'V'
                patterns. Defaults to None.

        Raises:
            ValueError: When duration is not positive
                or when delay is negative
        """
        measure = round(100000.0 * measure) / 100000.0
        duration = round(100000.0 * duration) / 100000.0
        delay = round(100000.0 * delay) / 100000.0
        if duration <= 0:
            raise ValueError("Duration is not positive " + str(duration))
        if delay < 0:
            raise ValueError("Delay is negative " + str(duration))
        if pattern not in slide_patterns:
            raise ValueError("Unknown slide pattern " + str(pattern))
        if pattern == "V" and reflect_position is None:
            raise Exception("Slide pattern 'V' is given " + "without reflection point")

        super().__init__(measure, start_position, NoteType.complete_slide)
        self.duration = duration
        self.end_position = end_position
        self.pattern = pattern
        self.delay = delay
        self.reflect_position = reflect_position


class TouchTapNote(SimaiNote):
    def __init__(
        self, measure: float, position: int, region: str, is_firework: bool = False
    ) -> None:
        measure = round(measure * 100000.0) / 100000.0

        super().__init__(measure, position, NoteType.touch_tap)
        self.is_firework = is_firework
        self.region = region


class TouchHoldNote(SimaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        region: str,
        duration: float,
        is_firework: bool = False,
    ) -> None:
        measure = round(measure * 100000.0) / 100000.0
        duration = round(duration * 100000.0) / 100000.0

        super().__init__(measure, position, NoteType.touch_hold)
        self.is_firework = is_firework
        self.region = region
        self.duration = duration


class BPM(Event):
    def __init__(self, measure: float, bpm: float) -> None:
        if bpm <= 0:
            raise ValueError("BPM is not positive " + str(bpm))

        measure = round(measure * 100000.0) / 100000.0

        super().__init__(measure, EventType.bpm)
        self.bpm = bpm


def slide_to_pattern_str(slide_note: SlideNote) -> str:
    pattern = slide_note.pattern
    if pattern != "V":
        return pattern

    if slide_note.reflect_position is None:
        raise Exception("Slide has 'V' pattern but no reflect position")

    return "V{}".format(slide_note.reflect_position + 1)


def pattern_from_int(
    pattern: int, start_position: int, end_position: int
) -> Tuple[str, Optional[int]]:
    top_list = [0, 1, 6, 7]
    inv_slide_dict = {v: k for k, v in slide_dict.items()}
    dict_result = inv_slide_dict.get(pattern)
    if dict_result is not None:
        return dict_result, None
    if pattern in [2, 3]:
        # Have I told you how much I hate the simai format?
        is_cw = pattern == 3
        distance = slide_distance(start_position, end_position, is_cw)
        if 0 < distance <= 3:
            return "^", None
        if distance == 0:
            if start_position in top_list and is_cw:
                return ">", None
            if start_position in top_list and not is_cw:
                return "<", None
            if start_position not in top_list and is_cw:
                return "<", None

            return ">", None
        if (start_position in top_list and is_cw) or not (
            start_position in top_list or is_cw
        ):
            return ">", None

        return "<", None
    if pattern in [11, 12]:
        if pattern == 11:
            reflect_position = start_position - 2
            if reflect_position < 0:
                reflect_position += 8
        else:
            reflect_position = start_position + 2
            if reflect_position > 7:
                reflect_position -= 8

        return "V", reflect_position

    raise ValueError(f"Unknown pattern: {pattern}")


def pattern_to_int(slide_note: SlideNote) -> int:
    pattern = slide_note.pattern
    top_list = [0, 1, 6, 7]

    dict_result = slide_dict.get(pattern)
    if dict_result is not None:
        return dict_result
    elif pattern == "^":
        is_cw = slide_is_cw(slide_note.position, slide_note.end_position)
        if is_cw:
            return 3
        else:
            return 2
    elif pattern == ">":
        is_top = slide_note.position in top_list
        if is_top:
            return 3
        else:
            return 2
    elif pattern == "<":
        is_top = slide_note.position in top_list
        if is_top:
            return 2
        else:
            return 3
    elif pattern == "V":
        if slide_note.reflect_position is None:
            raise ValueError("Slide pattern 'V' has no reflect position defined")

        is_cw = slide_is_cw(slide_note.position, slide_note.reflect_position)
        if is_cw:
            return 12
        else:
            return 11
    else:
        raise ValueError(f"Unknown slide pattern {pattern}")



import math
from typing import List, Union, Optional, Tuple
from fractions import Fraction
from multiprocessing import Pool, Event
import os

from ..event import NoteType
from .simainote import (
    TapNote,
    HoldNote,
    SlideNote,
    TouchTapNote,
    TouchHoldNote,
    BPM,
    slide_to_pattern_str,
)
from .simai_parser import parse_fragment

ABORT = None


def _lcm(a: int, b: int) -> int:
    return a * b // math.gcd(a, b)


def get_rest(
    current_measure: float,
    next_measure: float,
    after_next_measure: Optional[float] = None,
    current_divisor: Optional[int] = None,
    max_den: int = 1000,
) -> Tuple[int, int, int]:
    # Finds the amount of rest needed to get to the next measure
    # Returns a tuple (whole, divisor, amount)
    if next_measure < current_measure:
        raise ValueError("Current measure is greater than next.")
    if next_measure == current_measure:
        if current_divisor is None:
            return 0, 4, 0

        return 0, current_divisor, 0

    difference = math.modf(next_measure - current_measure)
    difference_frac = Fraction(difference[0]).limit_denominator(max_den)

    if current_divisor is not None:
        _lcm_divisor = _lcm(current_divisor, difference_frac.denominator)
        if _lcm_divisor == current_divisor and next_measure - current_measure < 1:
            divisor = current_divisor
            whole = 0
            amount = int(
                difference[1] * divisor
                + difference_frac.numerator * (divisor / difference_frac.denominator)
            )
            return whole, divisor, amount

    if after_next_measure is not None:
        if next_measure > after_next_measure:
            raise ValueError("After next measure is greater than next measure")

        difference_after = math.modf(after_next_measure - next_measure)
        difference_after_frac = Fraction(difference_after[0]).limit_denominator(max_den)
        _lcm_divisor_after = _lcm(
            difference_frac.denominator, difference_after_frac.denominator
        )
        if _lcm_divisor_after <= 64 and next_measure - current_measure < 1:
            divisor = _lcm_divisor_after
            whole = 0
            amount = int(
                difference[1] * divisor
                + difference_frac.numerator * (divisor / difference_frac.denominator)
            )
            return whole, divisor, amount

    whole = int(difference[1])
    amount = difference_frac.numerator
    divisor = difference_frac.denominator
    return whole, divisor, amount


def get_measure_divisor(measures: List[float], max_den: int = 1000) -> Optional[int]:
    """Accepts a list of measures that all belong to the same whole measure.
    Example: [1.23, 1.5, 1.67, 1.89]
    Returns a divisor that would fit the list upto an upper bound.
    Otherwise, returns None.

    Args:
        measures: List of measure that has the same whole measure.
        max_den: The maximum denominator when making a fraction.

    Returns:
        Returns the smallest divisor that would fit all the measure in the list.
        Or None, if no divisor found."""
    if len(measures) == 0:
        return None

    differences = []
    previous_measure = int(measures[0])
    measures.sort()
    for measure in measures:
        differences.append(math.modf(measure - previous_measure)[0])
        previous_measure = measure

    current__lcm = 1
    for difference in differences:
        divisor = Fraction(difference).limit_denominator(max_den).denominator
        current__lcm = _lcm(current__lcm, divisor)
        if current__lcm > 64:
            return None

    return current__lcm


def handle_tap(tap: TapNote, slides: List[SlideNote], counter: int) -> Tuple[str, int]:
    result = ""
    note_type = tap.note_type
    if note_type in [NoteType.tap, NoteType.break_tap, NoteType.ex_tap]:
        # Regular, break, and ex tap note
        if note_type == NoteType.break_tap:
            modifier_string = "b"
        elif note_type == NoteType.ex_tap:
            modifier_string = "x"
        else:
            modifier_string = ""

        if counter > 0:
            result += "/"

        result += "{}{}".format(tap.position + 1, modifier_string)
    elif note_type in [NoteType.star, NoteType.break_star, NoteType.ex_star]:
        produced_slides = [slide for slide in slides if slide.position == tap.position]
        if len(produced_slides) > 0:
            return "", counter

        # Adding $ would make a star note with no slides
        if note_type == NoteType.break_star:
            modifier_string = "b$"
        elif note_type == NoteType.ex_star:
            modifier_string = "x$"
        else:
            modifier_string = "$"

        if counter > 0:
            result += "/"
        result += "{}{}".format(tap.position + 1, modifier_string)

    counter += 1
    return result, counter


def handle_hold(hold: HoldNote, counter: int, max_den: int = 1000) -> Tuple[str, int]:
    result = ""
    frac = Fraction(hold.duration).limit_denominator(max_den * 2)
    if hold.note_type == NoteType.ex_hold:
        modifier_string = "hx"
    else:
        modifier_string = "h"

    if counter > 0:
        result += "/"

    result += "{}{}[{}:{}]".format(
        hold.position + 1, modifier_string, frac.denominator, frac.numerator
    )

    counter += 1
    return result, counter


def handle_touch_tap(touch: TouchTapNote, counter: int) -> Tuple[str, int]:
    result = ""
    if touch.is_firework:
        modifier_string = "f"
    else:
        modifier_string = ""

    if counter > 0:
        result += "/"

    result += "{}{}{}".format(touch.region, touch.position + 1, modifier_string)

    counter += 1
    return result, counter


def handle_touch_hold(
    touch: TouchHoldNote, counter: int, max_den: int = 1000
) -> Tuple[str, int]:
    result = ""
    frac = Fraction(touch.duration).limit_denominator(max_den * 2)
    if touch.is_firework:
        modifier_string = "hf"
    else:
        modifier_string = "h"

    if counter > 0:
        result += "/"

    result += "{}{}[{}:{}]".format(
        touch.region, modifier_string, frac.denominator, frac.numerator
    )

    counter += 1
    return result, counter


def handle_slide(
    slide: SlideNote,
    taps: List[TapNote],
    positions: List[int],
    bpm: float,
    counter: int,
    max_den: int = 1000,
) -> Tuple[str, int]:
    result = ""
    stars = [star for star in taps if star.position == slide.position]
    if counter > 0 and slide.position not in positions:
        result += "/"

    if slide.position not in positions:
        if len(stars) == 0:
            # No star
            modifier_string = "?"
        elif stars[0].note_type == NoteType.break_star:
            modifier_string = "b"
        elif stars[0].note_type == NoteType.ex_star:
            # Ex star
            modifier_string = "x"
        else:
            modifier_string = ""
    else:
        # Regular star
        modifier_string = ""

    if slide.position in positions:
        start_position = "*"
    else:
        start_position = str(slide.position + 1)

    pattern = slide_to_pattern_str(slide)
    if slide.delay != 0.25:
        if slide.delay > 0.0025:
            scale = 0.25 / slide.delay
        else:
            # There are no instant slides in Simai due to its
            # "unique" way of representing slide delays
            # So we just make a very fast slide
            scale = 100

        equivalent_bpm = round(bpm * scale * 10000.0) / 10000.0
        equivalent_duration = slide.duration * scale
        frac = Fraction(equivalent_duration).limit_denominator(max_den * 10)
        result += "{}{}{}{}[{:.2f}#{}:{}]".format(
            start_position,
            modifier_string,
            pattern,
            slide.end_position + 1,
            equivalent_bpm,
            frac.denominator,
            frac.numerator,
        )
    else:
        frac = Fraction(slide.duration).limit_denominator(max_den * 10)
        result += "{}{}{}{}[{}:{}]".format(
            start_position,
            modifier_string,
            pattern,
            slide.end_position + 1,
            frac.denominator,
            frac.numerator,
        )

    if slide.position not in positions:
        positions.append(slide.position)

    counter += 1
    return result, counter


def convert_to_fragment(
    events: List[Union[TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote, BPM]],
    current_bpm: float,
    divisor: Optional[int] = None,
    max_den: int = 1000,
) -> str:
    # Accepts a list of events that starts at the same measure
    bpms = [bpm for bpm in events if isinstance(bpm, BPM)]
    tap_notes = [note for note in events if isinstance(note, TapNote)]
    hold_notes = [note for note in events if isinstance(note, HoldNote)]
    touch_tap_notes = [note for note in events if isinstance(note, TouchTapNote)]
    touch_hold_notes = [note for note in events if isinstance(note, TouchHoldNote)]
    slide_notes = [note for note in events if isinstance(note, SlideNote)]
    slide_notes.sort(
        key=lambda sn: (
            sn.position,
            sn.end_position,
            sn.pattern,
        )
    )

    fragment = ""
    counter = 0
    if len(bpms) != 0:
        fragment += "({})".format(bpms[0].bpm)

    if divisor is not None:
        fragment += f"{{{divisor}}}"

    for tap_note in tap_notes:
        result = handle_tap(tap_note, slide_notes, counter)
        fragment += result[0]
        counter = result[1]

    for hold_note in hold_notes:
        result = handle_hold(hold_note, counter, max_den=max_den)
        fragment += result[0]
        counter = result[1]

    for touch_tap_note in touch_tap_notes:
        result = handle_touch_tap(touch_tap_note, counter)
        fragment += result[0]
        counter = result[1]

    for touch_hold_note in touch_hold_notes:
        result = handle_touch_hold(touch_hold_note, counter, max_den=max_den)
        fragment += result[0]
        counter = result[1]

    positions: List[int] = []
    for slide_note in slide_notes:
        result = handle_slide(
            slide_note, tap_notes, positions, current_bpm, counter, max_den=max_den
        )
        fragment += result[0]
        counter = result[1]

    return fragment


def _parse_init(event):
    global ABORT
    ABORT = event


def _parse_helper(fragment: str) -> List:
    global ABORT
    # Return an empty list when ABORT is set or the fragments is empty or "E"
    if ABORT.is_set() or len(fragment) == 0 or fragment == "E":
        return []

    try:
        parsed = parse_fragment(fragment)
    except Exception as e:
        # Abort all jobs
        ABORT.set()
        raise RuntimeError(f"Error parsing fragment {fragment}") from e

    return parsed


def parallel_parse_fragments(fragments: List[str]) -> list:
    _abort = Event()

    cpu_count = os.cpu_count()
    if cpu_count is None:
        cpu_count = 1

    chunksize = 1 + len(fragments) // cpu_count

    # Stop jobs when abort is set
    def fragment_iter():
        for fragment in fragments:
            if not _abort.is_set():
                yield fragment

    with Pool(processes=cpu_count, initializer=_parse_init, initargs=(_abort,)) as pool:
        result = pool.map(_parse_helper, fragment_iter(), chunksize)

    return result



from typing import Union, Callable, Sequence
import copy

from ..event import MaiNote, NoteType
from ..maisxt import (
    MaiSxt,
    HoldNote as SDTHoldNote,
    SlideStartNote as SDTSlideStartNote,
)
from ..maima2 import MaiMa2, TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote

ma2_slide_dict = {
    "SI_": 1,
    "SCL": 2,
    "SCR": 3,
    "SUL": 4,
    "SUR": 5,
    "SSL": 6,
    "SSR": 7,
    "SV_": 8,
    "SXL": 9,
    "SXR": 10,
    "SLL": 11,
    "SLR": 12,
    "SF_": 13,
}


def _default_touch_converter(
    sdt: MaiSxt, touch_note: Union[TouchTapNote, TouchHoldNote]
) -> None:
    if isinstance(touch_note, TouchTapNote) and touch_note.region == "C":
        sdt.add_tap(measure=touch_note.measure, position=0)
    elif isinstance(touch_note, TouchTapNote):
        sdt.add_tap(
            measure=touch_note.measure, position=touch_note.position
        )
    elif isinstance(touch_note, TouchHoldNote) and touch_note.region == "C":
        sdt.add_hold(
            measure=touch_note.measure,
            position=0,
            duration=touch_note.duration,
        )


def ma2_to_sdt(
    ma2: MaiMa2,
    touch_converter: Callable[
        [MaiSxt, Union[TouchTapNote, TouchHoldNote]], None
    ] = _default_touch_converter,
    convert_touch: bool = False,
) -> MaiSxt:
    initial_bpm = ma2.get_bpm(0)
    sdt = MaiSxt(bpm=initial_bpm)
    convert_notes(sdt, ma2.notes, touch_converter, convert_touch)
    sdt.notes.sort()

    equivalent_notes = []
    for note in sdt.notes:
        current_measure = note.measure
        current_time = ma2.measure_to_second(current_measure)
        scale = sdt.bpm / ma2.get_bpm(current_measure)

        note = copy.deepcopy(note)
        note.measure = sdt.second_to_measure(current_time)

        if isinstance(note, SDTHoldNote):
            note.duration = note.duration * scale
        elif isinstance(note, SDTSlideStartNote):
            note.duration = note.duration * scale
            note.delay = note.delay * scale

        equivalent_notes.append(note)

    sdt.notes = equivalent_notes
    return sdt


def convert_notes(
    sdt: MaiSxt,
    ma2_notes: Sequence[MaiNote],
    touch_converter: Callable[[MaiSxt, Union[TouchTapNote, TouchHoldNote]], None],
    convert_touch: bool,
) -> None:
    skipped_notes = 0
    for ma2_note in ma2_notes:
        note_type = ma2_note.note_type
        if isinstance(ma2_note, TapNote):
            is_break = note_type in [NoteType.break_tap, NoteType.break_tap]
            is_star = note_type in [NoteType.star, NoteType.break_star]
            sdt.add_tap(
                measure=ma2_note.measure,
                position=ma2_note.position,
                is_break=is_break,
                is_star=is_star,
            )
        elif isinstance(ma2_note, HoldNote):
            # Hold, and ex hold
            sdt.add_hold(
                measure=ma2_note.measure,
                position=ma2_note.position,
                duration=ma2_note.duration,
            )
        elif isinstance(ma2_note, SlideNote):
            # Complete slide
            # SDT slide durations include the delay unlike in ma2
            sdt.add_slide(
                measure=ma2_note.measure,
                start_position=ma2_note.position,
                end_position=ma2_note.end_position,
                duration=ma2_note.duration + ma2_note.delay,
                pattern=ma2_note.pattern,
                delay=ma2_note.delay,
            )
        elif isinstance(ma2_note, (TouchTapNote, TouchHoldNote)):
            # Touch tap, and touch hold
            if convert_touch:
                touch_converter(sdt, ma2_note)
            else:
                skipped_notes += 1
        else:
            print("Warning: Unknown note type {}".format(note_type))

    if skipped_notes > 0:
        print("Skipped {} touch note(s)".format(skipped_notes))



from typing import Sequence

from ..simai import (
    SimaiChart,
    pattern_from_int,
)
from ..maima2 import MaiMa2, TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote
from ..event import MaiNote, NoteType


def ma2_to_simai(ma2: MaiMa2) -> SimaiChart:
    simai_chart = SimaiChart()

    for bpm in ma2.bpms:
        simai_chart.set_bpm(bpm.measure, bpm.bpm)

    convert_notes(simai_chart, ma2.notes)

    return simai_chart


def convert_notes(simai_chart: SimaiChart, ma2_notes: Sequence[MaiNote]) -> None:
    for ma2_note in ma2_notes:
        note_type = ma2_note.note_type
        if isinstance(ma2_note, TapNote):
            is_break = note_type in [NoteType.break_tap, NoteType.break_star]
            is_ex = note_type in [NoteType.ex_tap, NoteType.ex_star]
            is_star = note_type in [
                NoteType.star,
                NoteType.break_star,
                NoteType.ex_star,
            ]
            simai_chart.add_tap(
                measure=ma2_note.measure,
                position=ma2_note.position,
                is_break=is_break,
                is_star=is_star,
                is_ex=is_ex,
            )
        elif isinstance(ma2_note, HoldNote):
            is_ex = note_type == NoteType.ex_hold
            simai_chart.add_hold(
                measure=ma2_note.measure,
                position=ma2_note.position,
                duration=ma2_note.duration,
                is_ex=is_ex,
            )
        elif isinstance(ma2_note, SlideNote):
            # Ma2 slide durations does not include the delay
            # like in simai
            pattern = pattern_from_int(
                ma2_note.pattern, ma2_note.position, ma2_note.end_position
            )
            simai_chart.add_slide(
                measure=ma2_note.measure,
                start_position=ma2_note.position,
                end_position=ma2_note.end_position,
                duration=ma2_note.duration,
                pattern=pattern[0],
                delay=ma2_note.delay,
                reflect_position=pattern[1],
            )
        elif isinstance(ma2_note, TouchTapNote):
            simai_chart.add_touch_tap(
                measure=ma2_note.measure,
                position=ma2_note.position,
                region=ma2_note.region,
                is_firework=ma2_note.is_firework,
            )
        elif isinstance(ma2_note, TouchHoldNote):
            simai_chart.add_touch_hold(
                measure=ma2_note.measure,
                position=ma2_note.position,
                region=ma2_note.region,
                duration=ma2_note.duration,
                is_firework=ma2_note.is_firework,
            )
        else:
            print("Warning: Unknown note type {}".format(note_type))



import enum


class EventType(enum.Enum):
    note = 0
    bpm = 1
    meter = 2


class Event:
    def __init__(self, measure: float, event_type: EventType) -> None:
        if measure < 0:
            raise ValueError("Measure is negative " + str(measure))

        self.measure = round(measure * 10000) / 10000
        self.event_type = event_type



import argparse
import copy
import os.path

from maiconverter.converter import sdt_to_ma2
from maiconverter.maima2 import MaiMa2, HoldNote, SlideNote, TouchHoldNote
from maiconverter.maisxt import MaiSxt
from maiconverter.tool import quantise


# noinspection PyShadowingNames
def main():
    parser = argparse.ArgumentParser("sxt to ma2 with bpm changes")
    parser.add_argument("input", type=str)
    parser.add_argument("conform", type=str)
    parser.add_argument("bpm", type=float)
    parser.add_argument("--offset", type=str)
    parser.add_argument("--resolution", type=int, default=384)
    parser.add_argument("--quantise", type=int, default=16)

    args = parser.parse_args()

    sdt = MaiSxt.open(args.input, args.bpm)
    ma2 = sdt_to_ma2(sdt, args.bpm)
    ma2.notes.sort(key=lambda note: note.measure)

    conform_ma2 = MaiMa2.open(args.conform)

    new_notes = []
    offset = 0.0 - conform_ma2.second_to_measure(
        ma2.measure_to_second(0.0, decrement=False), increment=False
    )
    for note in ma2.notes:
        current_measure = note.measure
        current_time = ma2.measure_to_second(current_measure, decrement=False)
        current_conform_measure = conform_ma2.second_to_measure(
            current_time, increment=False
        )

        scale = conform_ma2.get_bpm(
            current_conform_measure, decrement=False
        ) / ma2.get_bpm(current_measure, decrement=False)

        note = copy.deepcopy(note)
        note.measure = quantise(current_conform_measure + offset, args.quantise)
        if isinstance(note, (HoldNote, TouchHoldNote)):
            note.duration = quantise(scale * note.duration, args.quantise)
        elif isinstance(note, SlideNote):
            note.duration = quantise(scale * note.duration, args.quantise)
            note.delay = quantise(scale * note.delay, args.quantise)

        new_notes.append(note)

    ma2.notes = new_notes
    ma2.bpms = conform_ma2.bpms

    if args.offset is not None:
        ma2.offset(args.offset)

    filename, _ = os.path.splitext(args.input)
    with open(filename + "_conformed.ma2", "w", newline="\r\n") as out:
        out.write(ma2.export(resolution=args.resolution))

    print("saved to: " + filename + "_conformed.ma2")


if __name__ == "__main__":
    main()



import copy
from typing import Union, Callable, List

from ..event import SimaiNote, NoteType
from ..maisxt import (
    MaiSxt,
    HoldNote as SDTHoldNote,
    SlideStartNote as SDTSlideStartNote,
)
from ..simai import (
    SimaiChart,
    pattern_to_int,
    TapNote,
    HoldNote,
    SlideNote,
    TouchHoldNote,
    TouchTapNote,
)


def _default_touch_converter(
    sxt: MaiSxt, touch_note: Union[TouchTapNote, TouchHoldNote]
) -> None:
    if isinstance(touch_note, TouchTapNote) and touch_note.region == "C":
        sxt.add_tap(measure=touch_note.measure, position=0)
    elif isinstance(touch_note, TouchTapNote):
        sxt.add_tap(
            measure=touch_note.measure,
            position=touch_note.position,
        )
    elif isinstance(touch_note, TouchHoldNote) and touch_note.region == "C":
        sxt.add_hold(
            measure=touch_note.measure,
            position=0,
            duration=touch_note.duration,
        )


def simai_to_sdt(
    simai: SimaiChart,
    touch_converter: Callable[
        [MaiSxt, Union[TouchHoldNote, TouchTapNote]], None
    ] = _default_touch_converter,
    convert_touch: bool = False,
) -> MaiSxt:
    initial_bpm = simai.get_bpm(1.0)
    sdt = MaiSxt(initial_bpm)
    convert_notes(sdt, simai.notes, touch_converter, convert_touch)
    sdt.notes.sort()

    equivalent_notes = []
    for note in sdt.notes:
        current_measure = note.measure
        current_time = simai.measure_to_second(current_measure)
        scale = sdt.bpm / simai.get_bpm(current_measure)

        note = copy.deepcopy(note)
        note.measure = sdt.second_to_measure(current_time)

        if isinstance(note, SDTHoldNote):
            note.duration = note.duration * scale
        elif isinstance(note, SDTSlideStartNote):
            note.duration = note.duration * scale
            note.delay = note.delay * scale

        equivalent_notes.append(note)

    sdt.notes = equivalent_notes
    return sdt


def convert_notes(
    sxt: MaiSxt,
    simai_notes: List[SimaiNote],
    touch_converter: Callable[[MaiSxt, Union[TouchHoldNote, TouchTapNote]], None],
    convert_touch: bool,
) -> None:
    skipped_notes = 0
    for simai_note in simai_notes:
        note_type = simai_note.note_type
        if isinstance(simai_note, TapNote):
            is_break = note_type in [NoteType.break_tap, NoteType.break_star]
            is_star = note_type in [NoteType.star, NoteType.break_star]
            sxt.add_tap(
                measure=simai_note.measure,
                position=simai_note.position,
                is_break=is_break,
                is_star=is_star,
            )
        elif isinstance(simai_note, HoldNote):
            sxt.add_hold(
                measure=simai_note.measure,
                position=simai_note.position,
                duration=simai_note.duration,
            )
        elif isinstance(simai_note, SlideNote):
            # SDT slide duration include the delay
            # unlike in simai
            pattern = pattern_to_int(simai_note)
            sxt.add_slide(
                measure=simai_note.measure,
                start_position=simai_note.position,
                end_position=simai_note.end_position,
                duration=simai_note.duration + simai_note.delay,
                pattern=pattern,
                delay=simai_note.delay,
            )
        elif isinstance(simai_note, (TouchTapNote, TouchHoldNote)):
            # Touch tap and touch hold
            if convert_touch:
                touch_converter(sxt, simai_note)
            else:
                skipped_notes += 1
        else:
            print(f"Warning: Unknown note type {note_type}")

    if skipped_notes > 0:
        print(f"Skipped {skipped_notes} touch note(s)")



from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="MaiConverter",
    description="Parse and convert Maimai chart formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="donmai",
    url="https://github.com/donmai-me/MaiConverter",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Other Audience",
        "Topic :: Games/Entertainment :: Arcade",
    ],
    packages=[
        "maiconverter",
        "maiconverter.event",
        "maiconverter.maicrypt",
        "maiconverter.maima2",
        "maiconverter.maisxt",
        "maiconverter.simai",
        "maiconverter.converter",
        "maiconverter.tool",
    ],
    package_data={"": ["*.lark"]},
    entry_points={
        "console_scripts": ["maiconverter=maiconverter.cli:main"],
    },
    python_requires="~=3.8",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=["pycryptodome~=3.9", "lark-parser~=0.11"],
)



def slide_distance(start: int, end: int, is_cw: bool = True) -> int:
    """Returns the distance between the start and end buttons. Returns the clockwise distance by default
    unless is_cw is set to False, in which case the counter-clockwise distance is returned.

    Args:
        start: The button where the slide begins. Buttons start at 0.
        end: The button where the slide ends. Buttons start at 0.
        is_cw: Determines if the returned distance is in terms of clockwise or counter-clockwise rotation.
            Defaults to True.
    """
    if is_cw:
        if start >= end:
            end += 8

        return (end - start) % 8

    if start <= end:
        start += 8

    return (start - end) % 8


def slide_is_cw(start_position: int, end_position: int) -> bool:
    """Returns True when the shortest distance between the start and end position is achieved with
    a clockwise rotation. Returns False otherwise.

    Args:
        start_position: The button where the slide begins. Buttons start at 0.
        end_position: The button where the slide ends. Buttons start at 0.

    Raises:
        ValueError: When the distance is 4 places. This is because either rotation produces the same
            distance.
    """
    # Handles slide cases where the direction is not specified
    # Returns True for clockwise and False for counterclockwise
    diff = abs(end_position - start_position)
    other_diff = abs(8 - diff)
    if diff == 4:
        raise ValueError("Can't choose direction for a 180 degree angle.")

    if (end_position > start_position and diff > other_diff) or (
        end_position < start_position and diff < other_diff
    ):
        return False
    else:
        return True



from typing import List

from ..maima2 import MaiMa2
from ..simai import (
    SimaiChart,
    pattern_to_int,
    TapNote,
    HoldNote,
    SlideNote,
    TouchHoldNote,
    TouchTapNote,
)
from ..event import SimaiNote, NoteType


def simai_to_ma2(simai: SimaiChart, fes_mode: bool = False) -> MaiMa2:
    ma2 = MaiMa2(fes_mode=fes_mode)

    for bpm in simai.bpms:
        ma2.set_bpm(bpm.measure, bpm.bpm)

    ma2.set_meter(0.0, 4, 4)
    convert_notes(ma2, simai.notes)
    return ma2


def convert_notes(ma2: MaiMa2, simai_notes: List[SimaiNote]) -> None:
    for simai_note in simai_notes:
        note_type = simai_note.note_type
        if isinstance(simai_note, TapNote):
            is_break = note_type in [NoteType.break_tap, NoteType.break_star]
            is_ex = note_type in [NoteType.ex_tap, NoteType.ex_star]
            is_star = note_type in [
                NoteType.star,
                NoteType.break_star,
                NoteType.ex_star,
            ]
            ma2.add_tap(
                measure=simai_note.measure,
                position=simai_note.position,
                is_break=is_break,
                is_star=is_star,
                is_ex=is_ex,
            )
        elif isinstance(simai_note, HoldNote):
            is_ex = note_type == NoteType.ex_hold
            ma2.add_hold(
                measure=simai_note.measure,
                position=simai_note.position,
                duration=simai_note.duration,
                is_ex=is_ex,
            )
        elif isinstance(simai_note, SlideNote):
            # Ma2 slide durations does not include the delay
            # like in simai
            pattern = pattern_to_int(simai_note)
            ma2.add_slide(
                measure=simai_note.measure,
                start_position=simai_note.position,
                end_position=simai_note.end_position,
                duration=simai_note.duration,
                pattern=pattern,
                delay=simai_note.delay,
            )
        elif isinstance(simai_note, TouchTapNote):
            ma2.add_touch_tap(
                measure=simai_note.measure,
                position=simai_note.position,
                region=simai_note.region,
                is_firework=simai_note.is_firework,
            )
        elif isinstance(simai_note, TouchHoldNote):
            ma2.add_touch_hold(
                measure=simai_note.measure,
                position=simai_note.position,
                region=simai_note.region,
                duration=simai_note.duration,
                is_firework=simai_note.is_firework,
            )
        else:
            print(f"Warning: Unknown note type {note_type}")



from typing import List
from lark import Lark, Transformer


class SimaiTransformer(Transformer):
    def title(self, n):
        n = n[0]
        return {"type": "title", "value": n.rstrip()}

    def artist(self, n):
        n = n[0]
        return {"type": "artist", "value": n.rstrip()}

    def smsg(self, n):
        pass

    def des(self, n):
        pass

    def freemsg(self, n):
        pass

    def first(self, n):
        pass

    def pvstart(self, n):
        pass

    def pvend(self, n):
        pass

    def wholebpm(self, n):
        pass

    def level(self, n):
        num, level = n
        return {"type": "level", "value": (int(num), level.rstrip())}

    def chart(self, n):
        num, raw_chart = n
        chart = ""
        for x in raw_chart.splitlines():
            if "||" not in x:
                chart += x

        chart = "".join(chart.split())
        return {"type": "chart", "value": (int(num), chart)}

    def amsg_first(self, n):
        pass

    def amsg_time(self, n):
        pass

    def amsg_content(self, n):
        pass

    def demo_seek(self, n):
        pass

    def demo_len(self, n):
        pass

    def chain(self, values):
        result = []
        for value in values:
            if isinstance(value, dict):
                result.append(value)

        return result


class FragmentTransformer(Transformer):
    def bpm(self, n) -> dict:
        (n,) = n
        event_dict = {
            "type": "bpm",
            "value": float(n),
        }
        return event_dict

    def divisor(self, n) -> dict:
        (n,) = n
        if float(n) == 0:
            raise ValueError("Divisor is 0.")

        event_dict = {
            "type": "divisor",
            "value": float(n),
        }
        return event_dict

    def equivalent_bpm(self, n) -> dict:
        if len(n) == 0:
            return {"type": "equivalent_bpm", "value": None}

        (n,) = n
        return {"type": "equivalent_bpm", "value": float(n)}

    def duration(self, items) -> dict:
        # Set defaults
        equivalent_bpm = None
        den = None
        num = None

        for item in items:
            if isinstance(item, str) and item.type == "INT" and den is None:
                den = int(item)
                if den <= 0:
                    return {
                        "type": "duration",
                        "equivalent_bpm": equivalent_bpm,
                        "duration": 0,
                    }
            elif isinstance(item, str) and item.type == "INT" and num is None:
                num = int(item)
            elif isinstance(item, str) and item[-1] == "#":
                equivalent_bpm = float(item[:-1])

        if den is None or num is None:
            raise ValueError("No denominator or numerator given")

        return {
            "type": "duration",
            "equivalent_bpm": equivalent_bpm,
            "duration": num / den,
        }

    def slide_beg(self, items) -> dict:
        text = items[0]
        start = int(text[0]) - 1
        text = text[1:]

        modifier = ""
        for i, char in enumerate(text):
            if char in "b@x?$!":
                modifier += char
            else:
                text = text[i:]
                break

        reflect = None

        if text[0] not in "-^<>szvwpqV":
            raise ValueError(f"Unknown slide pattern {text[0]}")

        pattern = text[0]
        if text[0] in "-^<>szvw":
            text = text[1:]
        elif text[0] in "pq" and text[1] in "pq":
            pattern += text[1]
            text = text[2:]
        elif text[0] in "pq" and text[1] not in "pq":
            text = text[1:]
        else:  # V slide
            reflect = int(text[1]) - 1
            text = text[2:]

        end = int(text[0]) - 1

        return {
            "type": "slide_beg",
            "start": start,
            "modifier": modifier,
            "pattern": pattern,
            "reflect": reflect,
            "end": end,
        }

    def chained_slide_note(self, items) -> dict:
        (
            text,
            duration_dict,
        ) = items

        # Skip the modifiers in a chained slide
        for i, char in enumerate(text):
            if char not in "b@x?$!":
                text = text[i:]
                break

        pattern = text[0]
        reflect = None
        if text[0] in "-^<>szvw":
            text = text[1:]
        elif text[0] in "pq" and text[1] in "pq":
            pattern += text[1]
            text = text[2:]
        elif text[0] in "pq" and text[1] not in "pq":
            text = text[1:]
        else:  # V slide
            reflect = int(text[1]) - 1
            text = text[2:]

        end = int(text[0]) - 1

        if not isinstance(duration_dict, dict):
            raise ValueError(f"Not a dict: {duration_dict}")

        duration = duration_dict["duration"]
        equivalent_bpm = duration_dict["equivalent_bpm"]

        return {
            "type": "chained_slide_note",
            "pattern": pattern,
            "reflect": reflect,
            "end": end,
            "equivalent_bpm": equivalent_bpm,
            "duration": duration,
        }

    def slide_note(self, items):
        start = None
        end = None
        modifier = ""
        pattern = None
        reflect = None
        equivalent_bpm = None
        duration = None
        chained_slides = []

        for item in items:
            if isinstance(item, dict) and item["type"] == "slide_beg":
                if item["start"] == -1 or item["reflect"] == -1 or item["end"] == -1:
                    return

                start = item["start"]
                modifier = item["modifier"]
                pattern = item["pattern"]
                end = item["end"]
                reflect = item["reflect"]
            elif isinstance(item, dict) and item["type"] == "duration":
                equivalent_bpm = item["equivalent_bpm"]
                duration = item["duration"]
            elif isinstance(item, dict) and item["type"] == "chained_slide_note":
                chained_slides.append(item)
            else:
                raise ValueError(f"Unknown value: {item}")

        if any((c is None) for c in [start, end, pattern, duration]):
            raise ValueError("Incomplete data")

        slides = []
        if start != -1 and end != -1 and reflect != -1:
            note_dict = {
                "type": "slide",
                "start_button": start,
                "modifier": modifier,
                "pattern": pattern,
                "reflect_position": reflect,
                "end_button": end,
                "duration": duration,
                "equivalent_bpm": equivalent_bpm,
            }
            slides.append(note_dict)

        if len(chained_slides) != 0:
            slides += process_chained_slides(start, modifier + "*", chained_slides)

        if len(slides) > 0:
            return slides

    def tap_hold_note(self, items):
        if len(items) == 2:
            (
                text,
                duration_dict,
            ) = items
        else:
            (text,) = items
            duration_dict = None

        button = int(text[0]) - 1
        text = text[1:]
        if button == -1:
            # Ignore simai notes that has button position 0
            return

        is_tap = True
        if "h" in text:
            is_tap = False

        modifier = ""
        for char in text:
            if char == "h":
                continue

            if is_tap and char in "bx$":
                modifier += char
            elif not is_tap and char in "x":
                modifier += char

        if not is_tap:
            if duration_dict is None:
                duration = 0
            else:
                duration = duration_dict["duration"]

            return {
                "type": "hold",
                "button": button,
                "modifier": modifier,
                "duration": duration,
            }

        return {
            "type": "tap",
            "button": button,
            "modifier": modifier,
        }

    def touch_tap_hold_note(self, items):
        if len(items) == 2:
            (
                text,
                duration_dict,
            ) = items
        else:
            (text,) = items
            duration_dict = None

        region = text[0]
        if len(text) > 1 and text[1] in "012345678":
            position = int(text[1]) - 1
            text = text[2:]
        else:
            position = 0
            text = text[1:]

        if region not in "CBE" or position == -1:
            return

        is_tap = True
        if "h" in text:
            is_tap = False

        modifier = ""
        for char in text:
            if char == "h":
                continue

            if char in "f":
                modifier += char

        if not is_tap:
            if duration_dict is None:
                duration = 0
            else:
                duration = duration_dict["duration"]

            return {
                "type": "touch_hold",
                "region": region,
                "location": position,
                "modifier": modifier,
                "duration": duration,
            }

        return {
            "type": "touch_tap",
            "region": region,
            "location": position,
            "modifier": modifier,
        }

    def pseudo_each(self, items):
        (item,) = items
        if isinstance(item, list):
            notes = item
        elif isinstance(item, dict):
            notes = [item]
        else:
            raise TypeError(f"Invalid type: {type(item)}")

        for note in notes:
            note["modifier"] += "`"

        return notes

    def chain(self, items) -> list:
        result = []
        # Flatten list
        for item in items:
            if isinstance(item, list):
                for subitem in item:
                    result.append(subitem)
            elif isinstance(item, dict):
                result.append(item)
        return result


def process_chained_slides(start_button, slide_modifier, chained_slides):
    complete_slides = []
    for slide in chained_slides:
        if start_button == -1 or slide["reflect"] == -1 or slide["end"] == -1:
            continue

        note_dict = {
            "type": "slide",
            "start_button": start_button,
            "modifier": slide_modifier,
            "pattern": slide["pattern"],
            "reflect_position": slide["reflect"],
            "end_button": slide["end"],
            "duration": slide["duration"],
            "equivalent_bpm": slide["equivalent_bpm"],
        }
        complete_slides.append(note_dict)

    return complete_slides


def parse_fragment(fragment: str, lark_file: str = "simai_fragment.lark") -> List[dict]:
    parser = Lark.open(lark_file, rel_to=__file__, parser="lalr")
    try:
        return FragmentTransformer().transform(parser.parse(fragment))
    except Exception:
        print(f"Error parsing {fragment}")
        raise
